/* $Id: detmon.c,v 1.11 2013-07-19 12:00:24 jtaylor Exp $
 *
 * This file is part of the irplib package
 * Copyright (C) 2002, 2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02111-1307 USA
 */

/*
 * $Author: jtaylor $
 * $Date: 2013-07-19 12:00:24 $
 * $Revision: 1.11 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*---------------------------------------------------------------------------
                                  Includes
 ---------------------------------------------------------------------------*/

#include "detmon_darkron.h"

#include "detmon_ronbias.h"
#include "detmon.h"

#include "irplib_ksigma_clip.h"
#include "irplib_hist.h"
#include "irplib_utils.h"

#include <math.h>
#include <string.h>
#include <assert.h>
#include <float.h>

/**@{*/
/*--------------------------------------------------------------------------*/

/*
 * @defgroup detmon        Detector monitoring functions
 */

/*--------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
                                  Defines
 ---------------------------------------------------------------------------*/

/* Computes the square of an euclidean distance bet. 2 points */
#define pdist(x1,y1,x2,y2) (((x1-x2)*(x1-x2))+((y1-y2)*(y1-y2)))

#define cpl_drand() ((double)rand()/(double)RAND_MAX)

enum pixeltypes
{
    HOT = 0,
    DEAD = 1,
    NOISY = 2
};

enum stackingtypes
{
    MINMAX = 0,
    MEAN = 1,
    MEDIAN = 2,
    KSIGMA = 3
};

enum readouts
{
    HORIZONTAL = 1,
    VERTICAL = 2
};



static struct
{
    const char * ron_method;
    const char * dsnu_method;
    int         exts;
    int         nb_extensions;
    cpl_boolean opt_nir;
} detmon_dark_config;

#define NIR TRUE
#define OPT FALSE

/*---------------------------------------------------------------------------
  Private function prototypes
  ---------------------------------------------------------------------------*/

/* The following function is duplicated from cpl_det */
int
detmon_dark_dfs_set_groups(cpl_frameset *,
                           const char *);

cpl_error_code
detmon_dark_dsnu(cpl_frameset *,
                 cpl_imagelist *,
                 cpl_table *,
                 cpl_image *,
                 int pos);


static cpl_error_code
detmon_dark_save(const cpl_parameterlist *,
                 cpl_frameset *,
                 const char *,
                 const char *,
                 const char *,
                 const char *,
                 const char *,
                 const char *,
                 cpl_imagelist **,
                 cpl_table **,
                 cpl_imagelist **,
                 cpl_propertylist **,
                 const int,
                 const int,
                 const cpl_frameset *);



static cpl_error_code
detmon_retrieve_dark_params(const char *,
                            const char *,
                            const cpl_parameterlist *);

cpl_error_code
detmon_dark_qc(cpl_propertylist *,
               cpl_image *);



/*         RONBIAS FILLING PARLIST        */


/*---------------------------------------------------------------------------*/

/*
 * @brief  Fill input parameters with default values
 * @param  parlist      parameters list
 * @param  recipe_name  recipe name
 * @param  pipeline_name pipeline name

 * @return CPL_ERROR_NONE on success.
 */

/*---------------------------------------------------------------------------*/

    cpl_error_code
detmon_darkron_fill_parlist_default(cpl_parameterlist * parlist,
                                    const char *recipe_name,
                                    const char *pipeline_name)
{
    const cpl_error_code error =
        detmon_ronbias_fill_parlist(parlist, recipe_name, pipeline_name,
                                    "ALL", /* --method */
                                    "NORM",/* --pmethod */
                                    1,     /* --preoverscan_degree */
                                    -1,    /* --random_nsamples */
                                    -1,    /* --random_sizex */
                                    -1,    /* --random_sizey */
                                    0,     /* --criteria     */
                                    -1,    /* --ref_llx     */
                                    -1,    /* --ref_lly */
                                    -1,    /* --ref_urx */
                                    -1,    /* --ref_ury */
                                    "MEAN",/* --stacking_method */
                                    3,     /* --stacking_ks_low */
                                    3,     /* --stacking_ks_high */
                                    5,     /* --stacking_ks_iter */
                                    0,     /* --master_shift_x */
                                    0,     /* --master_shift_y */
                                    -1,    /* --ron_llx */
                                    -1,    /* --ron_lly */
                                    -1,    /* --ron_urx */
                                    -1,    /* --ron_ury */
                                    0,    /* --exts */
                                    NIR);
    cpl_ensure_code(!error, error);

    return cpl_error_get_code();
}


/*--------------------------------------------------------------------------*/

/*
 * @brief    Comparison function to identify different settings
 * @param    frame1  First frame
 * @param    frame2  Second frame
 * @return   0 if different, 1 if equal, -1 in error case
 */

/*--------------------------------------------------------------------------*/

int
detmon_compare_dits(const cpl_frame * frame1, const cpl_frame * frame2)
{
    int                     comparison;
    cpl_propertylist       *plist1;
    cpl_propertylist       *plist2;
    double                  dval1, dval2;

    /* Test entries */
    if(frame1 == NULL || frame2 == NULL)
        return -1;

    /* Get property lists */
    if((plist1 = cpl_propertylist_load(cpl_frame_get_filename(frame1),
                                       0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        return -1;
    }
    if((plist2 = cpl_propertylist_load(cpl_frame_get_filename(frame2),
                                       0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_propertylist_delete(plist1);
        return -1;
    }

    /* Test status */
    if(cpl_error_get_code()) {
        cpl_propertylist_delete(plist1);
        cpl_propertylist_delete(plist2);
        return -1;
    }

    /* Compare exposure time */
    comparison = 1;
    dval1 = irplib_pfits_get_exptime(plist1);
    dval2 = irplib_pfits_get_exptime(plist2);
    if(cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "cannot get exposure time");
        cpl_propertylist_delete(plist1);
        cpl_propertylist_delete(plist2);
        return -1;
    }
    if(fabs(dval1 - dval2) > 1e-3)
        comparison = 0;

    /* Free and return */
    cpl_propertylist_delete(plist1);
    cpl_propertylist_delete(plist2);
    return comparison;
}

/* End of duplicated code */


/*---------------------------------------------------------------------------*/
/*
 * @brief  dark recipe
 * @param  frameset       input frameset
 * @param  parlist        input parameter list
 * @param  tag            input recipe frame tag
 * @param  recipe_name    input recipe name
 * @param  pipeline_name  input pipeline name
 * @param  procatg_master input procatg of master product
 * @param  procatg_dsnu   input procatg dsnu product
 * @param  procatg_tbl    input procatg table
 * @param  package        input package name
 * @param  compare        function to make comparisons
 * @return CPL_ERROR_NONE on success.
 */
/*---------------------------------------------------------------------------*/


cpl_error_code
detmon_dark(cpl_frameset      * frameset,
            const cpl_parameterlist * parlist,
            const char        * tag,
            const char        * recipe_name,
            const char        * pipeline_name,
            const char        * procatg_master,
            const char        * procatg_dsnu,
            const char        * procatg_tbl,
            const char        * package,
            int              (*compare)(const cpl_frame *,
                                        const cpl_frame *))
{
    cpl_size                     nsets;
    cpl_size                    *selection = NULL;
    int                     i;
    cpl_error_code          error;

    if(detmon_dark_dfs_set_groups(frameset, tag)) {
        cpl_msg_error(cpl_func, "Cannot identify RAW and CALIB frames");
    }

    /*
     * This function reads all inputs parameters from parlist
     * and stores them in a global variable detmon_ronbias_config.
     * Similar to detmon_lg_retrieve_parlist(). See detmon.c
     */
    error = detmon_retrieve_dark_params(pipeline_name,
                                        recipe_name, parlist);
    cpl_ensure_code(!error, error);

    /* Labelise all input frames */
    if(compare == NULL)
        nsets = 1;
    else {
        cpl_msg_info(cpl_func, "Identify the different settings");
        selection = cpl_frameset_labelise(frameset, compare, &nsets);
        if(selection == NULL)
            cpl_msg_error(cpl_func, "Cannot labelise input frames");
    }

    detmon_dark_config.nb_extensions = 1;
    if(detmon_dark_config.exts < 0) {
        const cpl_frame        *cur_frame =
            cpl_frameset_get_position_const(frameset, 0);
        /* Get the nb of extensions */
        detmon_dark_config.nb_extensions =
            cpl_frame_get_nextensions(cur_frame);
    }

    /* Extract settings and reduce each of them */
    for(i = 0; i < nsets; i++) {
        cpl_size                    *select_dits = NULL;
        cpl_frameset           *cur_fset =
            nsets == 1 ? cpl_frameset_duplicate(frameset) :
            cpl_frameset_extract(frameset, selection, i);

        cpl_size                     ndits = 0;
        int                     j, k;
        cpl_table ** dsnu_table = NULL;
        cpl_imagelist ** dsnu = NULL;

        cpl_propertylist ** qclist =
            (cpl_propertylist **)
            cpl_malloc(detmon_dark_config.nb_extensions *
                       sizeof(cpl_propertylist *));


        cpl_imagelist ** masters =
            (cpl_imagelist **)
            cpl_malloc(detmon_dark_config.nb_extensions *
                       sizeof(cpl_imagelist *));

        /* Initialise memory for products */
        if(detmon_dark_config.opt_nir == OPT) {
            dsnu_table =
                (cpl_table **) cpl_malloc(detmon_dark_config.nb_extensions *
                                          sizeof(cpl_table *));
            dsnu =
                (cpl_imagelist **)
                cpl_malloc(detmon_dark_config.nb_extensions *
                           sizeof(cpl_imagelist *));
        }

        select_dits = cpl_frameset_labelise(cur_fset,
                                            detmon_compare_dits,
                                            &ndits);

        if(detmon_dark_config.exts >= 0) {
            *masters = cpl_imagelist_new();
            if(detmon_dark_config.opt_nir == OPT) {
                *dsnu = cpl_imagelist_new();
                *dsnu_table = cpl_table_new(ndits);
            }
            *qclist = cpl_propertylist_new();
            cpl_table_new_column(*dsnu_table, "DIT", CPL_TYPE_DOUBLE);
            cpl_table_new_column(*dsnu_table, "STDEV", CPL_TYPE_DOUBLE);
        } else {
            for ( j = 0; j < detmon_dark_config.nb_extensions; j ++) {
                masters[j] = cpl_imagelist_new();
                if(detmon_dark_config.opt_nir == OPT) {
                    dsnu[j] = cpl_imagelist_new();
                    dsnu_table[j] = cpl_table_new(ndits);
                }
                qclist[j] = cpl_propertylist_new();
                cpl_table_new_column(dsnu_table[j], "DIT", CPL_TYPE_DOUBLE);
                cpl_table_new_column(dsnu_table[j], "STDEV", CPL_TYPE_DOUBLE);
            }
        }

        for(j = 0; j < ndits; j++) {
            cpl_frameset  * cur_fdit = cpl_frameset_extract(cur_fset,
                                                            select_dits, j);
            cpl_imagelist ** raws =
                (cpl_imagelist **)
                cpl_malloc(detmon_dark_config.nb_extensions *
                           sizeof(cpl_imagelist *));

            if(detmon_dark_config.exts >= 0) {
                cpl_image * collapsed;
                *raws =
                    cpl_imagelist_load_frameset(cur_fdit, CPL_TYPE_FLOAT, 1,
                                                detmon_dark_config.exts);
                collapsed = cpl_imagelist_collapse_create(*raws);
                cpl_imagelist_set(*masters, collapsed, j);
                if(detmon_dark_config.opt_nir == OPT) {
                    detmon_dark_dsnu(cur_fdit, *dsnu, *dsnu_table,
                                     collapsed, j);
                }
                detmon_dark_qc(*qclist, collapsed);
            } else {
                cpl_imagelist          *raws_all_exts =
                    cpl_imagelist_load_frameset(cur_fdit, CPL_TYPE_FLOAT, 1,
                                                -1);
                for(k = 0; k < detmon_dark_config.nb_extensions; k++) {
                    int nframes = cpl_frameset_get_size(cur_fdit);
                    int h;
                    cpl_image * collapsed;
                    for(h = 0; h < nframes; h++) {
                        cpl_image              *image =
                            cpl_imagelist_unset(raws_all_exts,
                                                (detmon_dark_config.
                                                 nb_extensions - 1 - k) * h);
                        cpl_imagelist_set(raws[k], image, h);
                    }
                    collapsed = cpl_imagelist_collapse_create(raws[k]);
                    cpl_imagelist_set(masters[k],collapsed, j);
                    if(detmon_dark_config.opt_nir == OPT) {
                        detmon_dark_dsnu(cur_fdit, dsnu[k],
                                         dsnu_table[j], collapsed, j);
                    }
                    detmon_dark_qc(qclist[k], collapsed);
                }
            }

            cpl_frameset_delete(cur_fdit);
            for(k = 0; k < detmon_dark_config.nb_extensions; k++) {
                cpl_imagelist_delete(raws[k]);
            }
            cpl_free(raws);
        } /* end of loop (for) around different DIT values */

        cpl_frameset_delete(cur_fset);

        detmon_dark_save(parlist, frameset, recipe_name, pipeline_name,
                         procatg_master, procatg_tbl, procatg_dsnu,
                         package, masters, dsnu_table, dsnu, qclist,
                         0, 0, frameset);

        if(detmon_dark_config.opt_nir == OPT) {
            for(j = 0; j < detmon_dark_config.nb_extensions; j++) {
                cpl_table_delete(dsnu_table[j]);
                cpl_imagelist_delete(dsnu[j]);
            }
            cpl_free(dsnu_table);
            cpl_free(dsnu);
        }

        for(j = 0; j < detmon_dark_config.nb_extensions; j++) {
            cpl_propertylist_delete(qclist[j]);
            cpl_imagelist_delete(masters[j]);
        }
        cpl_free(qclist);
        cpl_free(masters);
        cpl_free(select_dits);

    } /* end of loop (for) around different setting */

    cpl_free(selection);

    return cpl_error_get_code();
}


/*---------------------------------------------------------------------------*/

/*
 * @brief  Classifies input frames
 * @param  set    Input frameset
 * @param  tag    Input frame tag

 * @return 0 on success, else -1
 */

/*---------------------------------------------------------------------------*/
int
detmon_dark_dfs_set_groups(cpl_frameset * set, const char *tag)
{



    /* Check entries */
    if(set == NULL)
        return -1;

    /* Initialize */
    int nframes = cpl_frameset_get_size(set);

    /* Loop on frames */
    for(int i = 0; i < nframes; i++) {
        cpl_frame* cur_frame = cpl_frameset_get_position(set, i);
        const char* cur_tag = cpl_frame_get_tag(cur_frame);

        /* RAW frames */
        if(!strcmp(cur_tag, tag))
            cpl_frame_set_group(cur_frame, CPL_FRAME_GROUP_RAW);
        /* CALIB frames */

        /*        else if (!strcmp(tag, IIINSTRUMENT_CALIB_FLAT))
                  cpl_frame_set_group(cur_frame, CPL_FRAME_GROUP_CALIB) ;
                  */
    }
    return 0;
}

/*---------------------------------------------------------------------------*/

/*
 * @brief  Retrieve input parameters
 * @param  pipeline_name        Input pipeline name
 * @param  recipe_name          Input recipe name
 * @param  parlist              Input parameter list
 * @return CPL_ERROR_NONE on success.
 */

/*---------------------------------------------------------------------------*/
static                  cpl_error_code
detmon_retrieve_dark_params(const char *pipeline_name,
                            const char *recipe_name,
                            const cpl_parameterlist * parlist)
{
    char                   *par_name;
    const cpl_parameter    *par;

    /* --ron.method */
    par_name = cpl_sprintf("%s.%s.ron.method", pipeline_name, recipe_name);
    assert(par_name != NULL);
    par = cpl_parameterlist_find_const(parlist, par_name);
    detmon_dark_config.ron_method = cpl_parameter_get_string(par);
    cpl_free(par_name);

    /* --dsnu.method */
    par_name = cpl_sprintf("%s.%s.dsnu.method", pipeline_name, recipe_name);
    assert(par_name != NULL);
    par = cpl_parameterlist_find_const(parlist, par_name);
    detmon_dark_config.dsnu_method = cpl_parameter_get_string(par);
    cpl_free(par_name);

    /* --opt_nir */
    par_name = cpl_sprintf("%s.%s.opt_nir", pipeline_name, recipe_name);
    assert(par_name != NULL);
    par = cpl_parameterlist_find_const(parlist, par_name);
    detmon_dark_config.opt_nir = cpl_parameter_get_bool(par);
    cpl_free(par_name);

    /* --exts */
    detmon_dark_config.exts =
        detmon_retrieve_par_int("exts", pipeline_name, recipe_name,
                                parlist);

    if(cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "Failed to retrieve the input parameters");
        cpl_ensure_code(0, CPL_ERROR_DATA_NOT_FOUND);
    }


    return CPL_ERROR_NONE;
}


/*---------------------------------------------------------------------------*/
/*
 * @brief  Fill dark param structure with default values
 * @param  parlist        input parameter list
 * @param  recipe_name    recipe name
 * @param  pipeline_name  pipeline name
 * @param  ron_method     method to compute RON
 * @param  dsnu_metthod   method to compute DSNU
 * @param  opt_nir        opt/nir swich parameter
 * @param  exts           extensions
 *
 * @return CPL_ERROR_NONE on success.
 */
/*---------------------------------------------------------------------------*/
    cpl_error_code
detmon_fill_dark_params(cpl_parameterlist * parlist,
                        const char *recipe_name,
                        const char *pipeline_name,
                        const char * ron_method,
                        const char * dsnu_method,
                        const char * opt_nir,
                        int exts)
{
    detmon_fill_parlist(parlist, recipe_name, pipeline_name, 4,

                        "ron.method",
                        "Method used to compute RON. Currently no "
                        "change is possible, RMS computed",
                        "CPL_TYPE_STRING", ron_method,

                        "dsnu.method",
                        "Method used to compute DSNU map. Currently no "
                        "change is possible. Method used STDEV",
                        "CPL_TYPE_STRING", dsnu_method,

                        "opt_nir",
                        "Boolean, OPT (FALSE) or NIR(TRUE)",
                        "CPL_TYPE_BOOL", opt_nir,

                        "exts",
                        "Activate the multi-exts option. Default 0"
                        "(primary unit), -1 (all exts)",
                        "CPL_TYPE_INT", exts);

    return cpl_error_get_code();
}

/*---------------------------------------------------------------------------*/
/*
 * @brief  Fill dark param structure with default values
 * @param  parlist        input parameter list
 * @param  recipe_name    recipe name
 * @param  pipeline_name  pipeline name
 *
 * @return CPL_ERROR_NONE on success.
 */
/*---------------------------------------------------------------------------*/
int
detmon_fill_dark_params_default(cpl_parameterlist * parlist,
                                const char *recipe_name,
                                const char *pipeline_name)
{
    detmon_fill_dark_params(parlist, recipe_name, pipeline_name,
                            "SIMPLE", /* --ron.method  */
                            "STDEV",  /* --dsnu.method */
                            "CPL_FALSE",  /* OPT*/
                            0);       /* --exts        */
    return cpl_error_get_code();
}

/*---------------------------------------------------------------------------*/

/*
 * @brief  Computes DSNU
 * @param  cur_fdit    Input frameset
 * @param  dsnu        Input imagelist
 * @param  dsnu_table  input dsnu table
 * @param  collapsed   input collapsed image
 * @param  pos         frame position in table
 * @return CPL_ERROR_NONE on success.
 */

/*---------------------------------------------------------------------------*/
cpl_error_code
detmon_dark_dsnu(cpl_frameset * cur_fdit,
                 cpl_imagelist * dsnu,
                 cpl_table * dsnu_table,
                 cpl_image * collapsed,
                 int pos)
{
    cpl_frame * first = cpl_frameset_get_position(cur_fdit, 0);
    cpl_propertylist * plist =
        cpl_propertylist_load(cpl_frame_get_filename(first), 0);
    double dit = irplib_pfits_get_exptime(plist);
    double mean = cpl_image_get_mean(collapsed);

    cpl_image * dsnu_map =
        cpl_image_subtract_scalar_create(collapsed, mean);
    double stdev;
    cpl_image_divide_scalar(dsnu_map, mean);
    stdev = cpl_image_get_stdev(dsnu_map);

    cpl_imagelist_set(dsnu, dsnu_map, pos);

    cpl_table_set(dsnu_table, "DIT", pos, dit);
    cpl_table_set(dsnu_table, "STDEV", pos, stdev);

    cpl_propertylist_delete(plist);

    return cpl_error_get_code();

}

/*---------------------------------------------------------------------------*/

/*
 * @brief  Save dark recipe products
 * @param  parlist        input parameter list
 * @param  frameset       input frameset
 * @param  recipe_name    input recipe name
 * @param  pipeline_name  input pipeline name
 * @param  procatg_master input procatg of master product
 * @param  procatg_tbl    input procatg table
 * @param  procatg_dsnu   input procatg dsnu product
 * @param  package        input package name
 * @param  masters        list of masters products
 * @param  dsnu_table     dsnu table product
 * @param  dsnu           list of dsnu images products
 * @param  qclist         qc parameters
 * @param  flag_sets      switch
 * @param  which_set      specifier
 * @param  usedframes     used frames
 * @return CPL_ERROR_NONE on success.
 */

/*---------------------------------------------------------------------------*/
static                  cpl_error_code
detmon_dark_save(const cpl_parameterlist * parlist,
                 cpl_frameset * frameset,
                 const char *recipe_name,
                 const char *pipeline_name,
                 const char *procatg_master,
                 const char *procatg_tbl,
                 const char *procatg_dsnu,
                 const char *package,
                 cpl_imagelist ** masters,
                 cpl_table ** dsnu_table,
                 cpl_imagelist ** dsnu,
                 cpl_propertylist ** qclist,
                 const int flag_sets,
                 const int which_set,
                 const cpl_frameset * usedframes)
{

    cpl_frame              *ref_frame;
    cpl_propertylist       *plist;
    char                   *name_o = NULL; /* Avoid (false) uninit warning */
    int                     i, j;
    cpl_propertylist       *paflist;
    cpl_error_code          error;
    int nb_images;

    /***************************/
    /*  Write the MASTER FITS  */
    /***************************/

    nb_images = cpl_imagelist_get_size(masters[0]);
    cpl_ensure_code(nb_images > 0, CPL_ERROR_DATA_NOT_FOUND);


    for(i = 0; i < nb_images; i++) {
        /* Set the file name for each image */
        if(!flag_sets) {
            name_o =
                cpl_sprintf("%s_master_dit_%d.fits", recipe_name, i+1);
            assert(name_o != NULL);
        } else {
            name_o =
                cpl_sprintf("%s_master_dit_%d_set%02d.fits",
                            recipe_name, i, which_set);
            assert(name_o != NULL);
        }


        /* Save the image */
        if(detmon_dark_config.exts >= 0) {
            cpl_propertylist * pro_master = cpl_propertylist_new();

            cpl_propertylist_append_string(pro_master,
                                           CPL_DFS_PRO_CATG, procatg_master);

            cpl_propertylist_append(pro_master, qclist[0]);

            if(cpl_dfs_save_image
               (frameset, NULL, parlist, usedframes, NULL,
                cpl_imagelist_get(*masters, i), CPL_BPP_IEEE_FLOAT,
                recipe_name, pro_master, NULL, package,
                name_o)) {
                cpl_msg_error(cpl_func, "Cannot save the product: %s",
                              name_o);
                cpl_free(name_o);
                cpl_ensure_code(0, CPL_ERROR_FILE_NOT_CREATED);

            }

            cpl_propertylist_delete(pro_master);
        } else {
            cpl_propertylist * pro_master = cpl_propertylist_new();

            cpl_propertylist_append_string(pro_master,
                                           CPL_DFS_PRO_CATG, procatg_master);

            cpl_propertylist_append(pro_master, qclist[0]);

            if(cpl_dfs_save_image(frameset, NULL, parlist, usedframes, NULL,
                                  NULL, CPL_BPP_IEEE_FLOAT, recipe_name,
                                  pro_master, NULL,
                                  package, name_o)) {
                cpl_msg_error(cpl_func, "Cannot save the product: %s",
                              name_o);
                cpl_free(name_o);
                cpl_ensure_code(0, CPL_ERROR_FILE_NOT_CREATED);
            }

            cpl_propertylist_delete(pro_master);
            for(j = 0; j < detmon_dark_config.nb_extensions; j++) {
                error =
                    cpl_image_save(cpl_imagelist_get(masters[j], i),
                                   name_o, CPL_BPP_IEEE_FLOAT, qclist[j],
                                   CPL_IO_EXTEND);
                cpl_ensure_code(!error, error);
            }
        }
        cpl_free(name_o);
    }

    if (detmon_dark_config.opt_nir == OPT) {
        cpl_propertylist * pro_tbl = cpl_propertylist_new();

        cpl_propertylist_append_string(pro_tbl,
                                       CPL_DFS_PRO_CATG, procatg_tbl);

        cpl_propertylist_append(pro_tbl, qclist[0]);
        /*******************************/
        /*  Write the LINEARITY TABLE  */
        /*******************************/

        /* Set the file name for the table */
        if(!flag_sets) {
            name_o = cpl_sprintf("%s_dsnu_table.fits", recipe_name);
            assert(name_o != NULL);
        } else {
            name_o =
                cpl_sprintf("%s_dsnu_table_set%02d.fits", recipe_name,
                            which_set);
            assert(name_o != NULL);
        }
        /* Save the table */
        if(cpl_dfs_save_table(frameset, NULL, parlist, usedframes, NULL,
                              dsnu_table[0], NULL, recipe_name, pro_tbl, NULL,
                              package, name_o)) {
            cpl_msg_error(cpl_func, "Cannot save the product: %s", name_o);
            cpl_free(name_o);
            cpl_ensure_code(0, CPL_ERROR_FILE_NOT_CREATED);
        }

        cpl_propertylist_delete(pro_tbl);

        if(detmon_dark_config.exts < 0) {

            for(i = 1; i < detmon_dark_config.nb_extensions; i++) {
                error =
                    cpl_table_save(dsnu_table[i], NULL, qclist[i], name_o,
                                   CPL_IO_EXTEND);
                cpl_ensure_code(!error, error);
            }
        }

        /* Free */
        cpl_free(name_o);

        /***************************/
        /*  Write the DSNU_MAP FITS  */
        /***************************/

        for(i = 0; i < nb_images; i++) {
            /* Set the file name for each image */
            if(!flag_sets) {
                name_o =
                    cpl_sprintf("%s_dsnu_map_dit_%d.fits", recipe_name, i+1);
                assert(name_o != NULL);
            } else {
                name_o =
                    cpl_sprintf("%s_dsnu_map_dit_%d_set%02d.fits",
                                recipe_name, i, which_set);
                assert(name_o != NULL);
            }


            /* Save the image */
            if(detmon_dark_config.exts >= 0) {
                cpl_propertylist * pro_dsnu = cpl_propertylist_new();

                cpl_propertylist_append_string(pro_dsnu,
                                               CPL_DFS_PRO_CATG, procatg_dsnu);

                cpl_propertylist_append(pro_dsnu, qclist[0]);

                if(cpl_dfs_save_image
                   (frameset, NULL, parlist, usedframes, NULL,
                    cpl_imagelist_get(*dsnu, i), CPL_BPP_IEEE_FLOAT,
                    recipe_name, pro_dsnu, NULL, package,
                    name_o)) {
                    cpl_msg_error(cpl_func, "Cannot save the product: %s",
                                  name_o);
                    cpl_free(name_o);
                    cpl_ensure_code(0, CPL_ERROR_FILE_NOT_CREATED);

                }

                cpl_propertylist_delete(pro_dsnu);
            } else {
                cpl_propertylist * pro_dsnu = cpl_propertylist_new();

                cpl_propertylist_append_string(pro_dsnu,
                                               CPL_DFS_PRO_CATG, procatg_dsnu);

                cpl_propertylist_append(pro_dsnu, qclist[0]);

                if(cpl_dfs_save_image(frameset, NULL, parlist, usedframes,
                                      NULL, NULL,
                                      CPL_BPP_IEEE_FLOAT, recipe_name,
                                      pro_dsnu, NULL,
                                      package, name_o)) {
                    cpl_msg_error(cpl_func, "Cannot save the product: %s",
                                  name_o);
                    cpl_free(name_o);
                    cpl_ensure_code(0, CPL_ERROR_FILE_NOT_CREATED);
                }

                cpl_propertylist_delete(pro_dsnu);
                for(j = 0; j < detmon_dark_config.nb_extensions; j++) {
                    error =
                        cpl_image_save(cpl_imagelist_get(dsnu[j], i),
                                       name_o, CPL_BPP_IEEE_FLOAT, qclist[j],
                                       CPL_IO_EXTEND);
                    cpl_ensure_code(!error, error);
                }
            }
            cpl_free(name_o);
        }



    } /* End of if(OPT) */

    /*******************************/
    /*  Write the PAF file(s)      */
    /*******************************/

    /* Get FITS header from reference file */
    ref_frame = cpl_frameset_get_position(frameset, 0);
    if((plist = cpl_propertylist_load(cpl_frame_get_filename(ref_frame),
                                      0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_ensure_code(0, cpl_error_get_code());
    }

    /* Get the keywords for the paf file */
    paflist = cpl_propertylist_new();
    cpl_propertylist_copy_property_regexp(paflist, plist,
                                          "^(ARCFILE|MJD-OBS|ESO TPL ID|"
                                          "DATE-OBS|ESO DET DIT|ESO DET NDIT|"
                                          "ESO DET NCORRS|"
                                          "ESO DET MODE NAME)$", 0);

    for(i = 0; i < detmon_dark_config.nb_extensions; i++) {
        cpl_propertylist * c_paflist = cpl_propertylist_duplicate(paflist);
        error = cpl_propertylist_append(c_paflist, qclist[i]);
        cpl_ensure_code(!error, error);

        /* Set the file name for the bpm */
        if(detmon_dark_config.exts >= 0) {
            if(!flag_sets) {
                name_o = cpl_sprintf("%s.paf", recipe_name);
                assert(name_o != NULL);
            } else {
                name_o = cpl_sprintf("%s_set%02d.paf", recipe_name, which_set);
                assert(name_o != NULL);
            }
        } else {
            if(!flag_sets) {
                name_o = cpl_sprintf("%s_ext%02d.paf", recipe_name, i+1);
                assert(name_o != NULL);
            } else {
                name_o = cpl_sprintf("%s_set%02d_ext%02d.paf", recipe_name, which_set, i+1);
                assert(name_o != NULL);
            }
        }
        /* Save the PAF */
        if(cpl_dfs_save_paf(pipeline_name, recipe_name, c_paflist, name_o)) {
            cpl_msg_error(cpl_func, "Cannot save the product: %s", name_o);
            cpl_free(name_o);
            cpl_propertylist_delete(paflist);
            cpl_propertylist_delete(plist);
            cpl_free(name_o);
            cpl_ensure_code(0, CPL_ERROR_FILE_NOT_CREATED);
        }
        cpl_propertylist_delete(c_paflist);
        cpl_free(name_o);
    }

    cpl_propertylist_delete(plist);
    cpl_propertylist_delete(paflist);

    return cpl_error_get_code();
}
/*---------------------------------------------------------------------------*/
/**
  @brief    Computes QC on dark frame
  @param    qclist    the input quality control propertylist to store results
  @param    collapsed the input image
  @return CPL_ERRR_NONE if success
  */
/*---------------------------------------------------------------------------*/
cpl_error_code
detmon_dark_qc(cpl_propertylist * qclist,
               cpl_image * collapsed)
{
    double mean = cpl_image_get_mean(collapsed);
    double stdev = cpl_image_get_stdev(collapsed);


    cpl_propertylist_append_double(qclist,DETMON_QC_DARK, mean);
    cpl_propertylist_set_comment(qclist,DETMON_QC_DARK,
                                         DETMON_QC_DARK_C);

    cpl_propertylist_append_double(qclist,DETMON_QC_DARK_STDEV, stdev);
    cpl_propertylist_set_comment(qclist,DETMON_QC_DARK_STDEV,
                                         DETMON_QC_DARK_STDEV_C);
    detmon_print_rec_status();

    return cpl_error_get_code();
}


/**@}*/
