/*
 * This file is part of the DETMON Pipeline
 * Copyright (C) 2013 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: jtaylor $
 * $Date: 2013-08-07 09:32:50 $
 * $Revision: 1.6 $
 * $Name: not supported by cvs2svn $
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*----------------------------------------------------------------------------
                             Includes
 ----------------------------------------------------------------------------*/


#include "detmon_lg.h"
#include "detmon.h"

#include "irplib_plugin.h"
#include "irplib_utils.h"

#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <assert.h>

/*----------------------------------------------------------------------------
				Private Functions prototypes
 ----------------------------------------------------------------------------*/
int
detmon_lg_dfs_set_groups(cpl_frameset * set,
                         const char *tag_on, const char *tag_off);

static int check_regions(const cpl_size *, const cpl_size *,
                         const cpl_size *, const cpl_size *,
                         cpl_size) CPL_ATTR_NONNULL;

static int compare_tags(const cpl_frame *, const cpl_frame *) CPL_ATTR_NONNULL;

static const cpl_parameter *
get_parameter_const(const cpl_parameterlist * parlist,
                    const char * recipe, const char * name)
{
    char * parname = cpl_sprintf("detmon.%s.%s", recipe, name);
    const cpl_parameter * p = cpl_parameterlist_find_const(parlist, parname);
    cpl_free(parname);
    return p;
}

static cpl_parameter *
get_parameter(cpl_parameterlist * parlist,
              const char * recipe, const char * name)
{
    char * parname = cpl_sprintf("detmon.%s.%s", recipe, name);
    cpl_parameter * p = cpl_parameterlist_find(parlist, parname);
    cpl_free(parname);
    return p;
}

static int compare_tags(const cpl_frame * a, const cpl_frame * b)
{
    return strcmp(cpl_frame_get_tag(a), cpl_frame_get_tag(b)) == 0;
}

/**@{*/

/* ---------------------------------------------------------------------------*/
/**
 * @brief merge qc keys into a property list, adding region subkey
 *
 * @param mplist    list to merge into
 * @param plist_    list to merge from
 * @param pos       region position, used as REG%d
 * @param rmregexp  regexp of keys to remove
 *
 * @return cpl_error_code
 */
/* ---------------------------------------------------------------------------*/
static cpl_error_code merge_keys(cpl_propertylist * mplist,
                                 const cpl_propertylist * plist_,
                                 const int pos,
                                 const char * rmregexp)
{
    char buffer[256];
    /* no copy_regex, so duplicate and use erase_regex on the copy */
    cpl_propertylist * plist = cpl_propertylist_duplicate(plist_);

    if (rmregexp) {
        cpl_propertylist_erase_regexp(plist, rmregexp, 0);
        cpl_propertylist_erase_regexp(mplist, rmregexp, 0);
    }

    for (cpl_size i = 0; i < cpl_propertylist_get_size(plist); i++) {
        cpl_property * p = cpl_propertylist_get(plist, i);
        const char * key = cpl_property_get_name(p);
        /* extend all qc keywords by a region prefix */
        if (!strncmp(key, "ESO QC ", 7)) {
            const char * com;
            sprintf(buffer, "ESO QC REG%d %s", pos, key + 7);
            cpl_msg_debug(cpl_func, "merging %s to %s", key, buffer);
            if (cpl_property_get_type(p) == CPL_TYPE_INT) {
                int val = cpl_property_get_int(p);
                cpl_propertylist_append_int(mplist, buffer, val);
            } else if (cpl_property_get_type(p) == CPL_TYPE_DOUBLE) {
                double val = cpl_property_get_double(p);
                cpl_propertylist_append_double(mplist, buffer, val);
            } else if (cpl_property_get_type(p) == CPL_TYPE_STRING) {
                const char * val = cpl_property_get_string(p);
                cpl_propertylist_append_string(mplist, buffer, val);
            }
            else
                bug_if(1);
            com = cpl_property_get_comment(p);
            cpl_propertylist_set_comment(mplist, buffer, com);

            /* remove unprefixed key */
            if (cpl_propertylist_has(mplist, key))
                cpl_propertylist_erase(mplist, key);
        }
    }
    skip_if(0);

    end_skip;
    cpl_propertylist_delete(plist);

    return cpl_error_get_code();
}


/* ---------------------------------------------------------------------------*/
/**
 * @brief  merge imagelists into another
 *
 * @param mimgl imagelist to merge into
 * @param img imagelist to merge from
 * @note images are merged by adding them
 */
/* ---------------------------------------------------------------------------*/
static cpl_error_code
merge_images(cpl_imagelist * mimgl, const cpl_imagelist * img)
{
    cpl_ensure_code(cpl_imagelist_get_size(img) ==
                    cpl_imagelist_get_size(mimgl), CPL_ERROR_ILLEGAL_INPUT);
    for (int i = 0; i < cpl_imagelist_get_size(img); i++) {
        cpl_image * mimg = cpl_imagelist_get(mimgl, i);
        const cpl_image * part = cpl_imagelist_get_const(img, i);
        cpl_image_add(mimg, part);
    }

    return cpl_error_get_code();
}


/* ---------------------------------------------------------------------------*/
/**
 * @brief  merge tables into another
 *
 * @param mtab table to merge into
 * @param tab  table to merge from
 * @param pos  region position, added as REGION integer column
 */
/* ---------------------------------------------------------------------------*/
static cpl_error_code
merge_table(cpl_table * mtab, cpl_table * tab, const int pos)
{
    cpl_table_new_column(tab, "REGION", CPL_TYPE_INT);
    cpl_table_fill_column_window(tab, "REGION", 0, cpl_table_get_nrow(tab), pos);
    skip_if(cpl_table_insert(mtab, tab, cpl_table_get_nrow(mtab)));

    end_skip;

    return cpl_error_get_code();
}


static cpl_error_code
parse_token(const char * token, cpl_size * llx, cpl_size * lly,
            cpl_size * urx, cpl_size * ury)
{
    assert(sizeof(*llx) == 8);
    if (sscanf(token, " %lld , %lld , %lld , %lld ", llx, lly, urx, ury) == 4) {
        return CPL_ERROR_NONE;
    }
    return cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT,
                                 "Data not formated correctly: %s; "
                                 "expecting: name,llx,lly,urx,ury",
                                 token);
}


/* ---------------------------------------------------------------------------*/
/**
 * @brief read region points from file
 *
 * @param filename file to read from
 * @param rllx pointer to array for lower left point
 * @param rlly pointer to array for lower left point
 * @param rurx pointer to array for upper right point
 * @param rury pointer to array for upper right point
 * @param n    number of regions read
 * @note  file is expected to contain the four points comma separated
 *        one group of four points per line
 */
/* ---------------------------------------------------------------------------*/
static cpl_error_code
read_regions_file(const char * filename, cpl_size ** rllx,
                  cpl_size ** rlly, cpl_size ** rurx, cpl_size ** rury,
                  cpl_size * n)
{
    FILE * f = fopen(filename, "r");
    char line[1024];
    cpl_size i = 0;
    cpl_size nmax = 100;
    if (f == NULL) {
        return cpl_error_set_message(cpl_func, CPL_ERROR_FILE_IO,
                                     "Could not open file %s: %s", filename,
                                     strerror(errno));
    }
    cpl_msg_info(cpl_func, "Reading %s", filename);

    *n = 0;
    *rllx = cpl_malloc(sizeof(**rllx) * nmax);
    *rlly = cpl_malloc(sizeof(**rllx) * nmax);
    *rurx = cpl_malloc(sizeof(**rllx) * nmax);
    *rury = cpl_malloc(sizeof(**rllx) * nmax);

    while (fgets(line, 1024, f) != NULL) {
        const size_t nline = strlen(line);
        if (line[nline - 1] != '\n') {
            cpl_error_set_message(cpl_func, CPL_ERROR_FILE_IO,
                                  "File %s contained too long line",
                                  filename);
            skip_if(0);
        }

        if (line[0] == '#')
            continue;

        skip_if(parse_token(line, &(*rllx)[i], &(*rlly)[i],
                            &(*rurx)[i], &(*rury)[i]));
        *n = ++i;
        if (i >= nmax) {
            nmax *= 2;
            *rllx = cpl_realloc(*rllx, sizeof(**rllx) * nmax);
            *rlly = cpl_realloc(*rlly, sizeof(**rllx) * nmax);
            *rurx = cpl_realloc(*rurx, sizeof(**rllx) * nmax);
            *rury = cpl_realloc(*rury, sizeof(**rllx) * nmax);
            skip_if(0);
        }
    }

    end_skip;
    if (f)
        fclose(f);
    return CPL_ERROR_NONE;
}


/* ---------------------------------------------------------------------------*/
/**
 * @brief read region points from string
 *
 * @param istr  string to read from
 * @param rllx pointer to array for lower left point
 * @param rlly pointer to array for lower left point
 * @param rurx pointer to array for upper right point
 * @param rury pointer to array for upper right point
 * @param n    number of regions read
 * @note  string is expected to countain foud comma separated points
 *        one group of four points separated by a colon
 */
/* ---------------------------------------------------------------------------*/
static cpl_error_code
read_regions_str(const char * istr, cpl_size ** rllx,
                  cpl_size ** rlly, cpl_size ** rurx, cpl_size ** rury,
                  cpl_size * n)
{
    cpl_size nmax = 100;
    char * str_ = cpl_strdup(istr);
    char * str = str_;
    *n = 0;
    *rllx = cpl_malloc(sizeof(**rllx) * nmax);
    *rlly = cpl_malloc(sizeof(**rllx) * nmax);
    *rurx = cpl_malloc(sizeof(**rllx) * nmax);
    *rury = cpl_malloc(sizeof(**rllx) * nmax);

    for (int i = 0; ;i++) {
        const char * token = strtok(str, ":");
        if (token == NULL)
            break;

        skip_if(parse_token(token, &(*rllx)[i], &(*rlly)[i],
                            &(*rurx)[i], &(*rury)[i]));
        (*n)++;
        if (i >= nmax - 1) {
            nmax *= 2;
            *rllx = cpl_realloc(*rllx, sizeof(**rllx) * nmax);
            *rlly = cpl_realloc(*rlly, sizeof(**rllx) * nmax);
            *rurx = cpl_realloc(*rurx, sizeof(**rllx) * nmax);
            *rury = cpl_realloc(*rury, sizeof(**rllx) * nmax);
            skip_if(0);
        }
        str = NULL;
    }

    end_skip;
    cpl_free(str_);

    return CPL_ERROR_NONE;
}

static int check_regions(const cpl_size * llx, const cpl_size * lly,
                         const cpl_size * urx, const cpl_size * ury,
                         cpl_size nregions)
{
    for (cpl_size i = 0; i < nregions; i++) {
        if (llx[i] < 1 || lly[i] < 1 ||
            urx[i] < 1 || ury[i] < 1) {
            return cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT,
                 "Region %lld (%lld,%lld,%lld,%lld) has entry smaller one",
                    i, llx[i], lly[i], urx[i], ury[i]);
        }
        if ((urx[i] - llx[i] < 1) || (ury[i] - lly[i] < 1)) {
            return cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT,
                 "Region %lld (%lld,%lld,%lld,%lld) has size zero",
                    i, llx[i], lly[i], urx[i], ury[i]);
        }
        for (cpl_size j = i + 1; j < nregions; j++) {
            if (!(llx[i] > urx[j] || urx[i] < llx[j] ||
                  lly[i] > ury[j] || ury[i] < lly[j])) {
                return cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT,
                    "Regions %lld (%lld,%lld,%lld,%lld) and "
                    "%lld (%lld,%lld,%lld,%lld) overlap (llx,lly,urx,ury)",
                    i, llx[i], lly[i], urx[i], ury[i],
                    j, llx[j], lly[j], urx[j], ury[j]);
            }
        }
    }

    return CPL_ERROR_NONE;
}


/* ---------------------------------------------------------------------------*/
/**
 * @brief merge linearity and gain tables
 *
 * @param mplist[]  array propertylist to merge into, must have size of #ext
 * @param mtab[]    array tables to merge into, must have size of #ext
 * @param fn        filename to read mergee
 * @param next      number of extensions
 * @param rmregexp  regex to remove keys from merged plist
 * @param pos       region position parameter
 *
 * @return cpl_error_code
 */
/* ---------------------------------------------------------------------------*/
static cpl_error_code
merge_lingain(cpl_propertylist * mplist[], cpl_table * mtab[],
              const char * fn, const cpl_size next,
              const char * rmregexp, int pos)
{
    cpl_errorstate cleanstate = cpl_errorstate_get();

    {
        cpl_propertylist * plist = cpl_propertylist_load(fn, 0);
        skip_if(plist == NULL);
        if (mplist[0] == NULL) {
            mplist[0] = plist;
            merge_keys(mplist[0], plist, pos, rmregexp);
        }
        else {
            merge_keys(mplist[0], plist, pos, rmregexp);
            cpl_propertylist_delete(plist);
        }
    }
    skip_if(0);

    for (int e = 1; e < next + 1; e++) {
        cpl_propertylist * plist = cpl_propertylist_load(fn, e);
        cpl_table * tab = cpl_table_load(fn, e, 0);
        /* empty extensions may exist */
        if (tab == NULL) {
            cpl_errorstate_set(cleanstate);
            cpl_propertylist_delete(plist);
            continue;
        }
        skip_if(plist == NULL);

        if (mplist[e] == NULL) {
            merge_keys(plist, plist, pos, rmregexp);
            cpl_table_new_column(tab, "REGION", CPL_TYPE_INT);
            cpl_table_fill_column_window(tab, "REGION", 0,
                                         cpl_table_get_nrow(tab), 1);
            skip_if(0);
            mplist[e] = plist;
            mtab[e] = tab;
        }
        else {
            merge_keys(mplist[e], plist, pos, rmregexp);
            merge_table(mtab[e], tab, pos);
            skip_if(0);
            cpl_propertylist_delete(plist);
            cpl_table_delete(tab);
        }
        skip_if(0);
    }

    end_skip;

    return cpl_error_get_code();
}


/* ---------------------------------------------------------------------------*/
/**
 * @brief merge bad pixel and coefficient images
 *
 * @param mplist[]  array propertylist to merge into, must have size of #ext
 * @param mimg[]    array images to merge into, must have size of #ext
 * @param fn        filename to read mergee
 * @param next      number of extensions
 * @param rmregexp  regex to remove keys from merged plist
 * @param pos       region position parameter
 *
 * @return cpl_error_code
 */
/* ---------------------------------------------------------------------------*/
static cpl_error_code
merge_bpmcoeff(cpl_propertylist * mplist[], cpl_imagelist * mimg[],
              const char * fn, const cpl_size next,
              const char * rmregexp, int pos)
{
    cpl_errorstate cleanstate = cpl_errorstate_get();

    for (int e = 0; e < next + 1; e++) {
        cpl_propertylist * plist = cpl_propertylist_load(fn, e);
        cpl_imagelist * img = cpl_imagelist_load(fn, CPL_TYPE_UNSPECIFIED, e);
        /* empty extensions may exist */
        if (img == NULL) {
            cpl_errorstate_set(cleanstate);
            cpl_propertylist_delete(plist);
            continue;
        }
        skip_if(plist == NULL);

        if (mimg[e] == NULL) {
            merge_keys(plist, plist, pos, rmregexp);
            mplist[e] = plist;
            mimg[e] = img;
        }
        else {
            merge_keys(mplist[e], plist, pos, rmregexp);
            merge_images(mimg[e], img);
            skip_if(0);
            cpl_propertylist_delete(plist);
            cpl_imagelist_delete(img);
        }
        skip_if(0);
    }

    end_skip;

    return cpl_error_get_code();
}

typedef enum {
    DETMON_BP_MAP_NL_IDX = 0,
    DETMON_LIN_INFO_IDX = 1,
    DETMON_GAIN_INFO_IDX = 2,
    DETMON_BPM_IDX = 3,
    DETMON_COEFF_IDX = 4,
    DETMON_MAX_IDX,
} detmon_idx;

static cpl_error_code disable_contamination(cpl_parameterlist * parlist,
                                            const char * recipe)
{
    /* "disable" by setting it to a region of size 1 */
    cpl_parameter * par;
    for (int i = 1; i < 6; i++) {
        char buffer[100];
        sprintf(buffer, "detmon.%s.llx%d", recipe, i);
        par = cpl_parameterlist_find(parlist, buffer);
        cpl_parameter_set_int(par, 1);
        sprintf(buffer, "detmon.%s.lly%d", recipe, i);
        par = cpl_parameterlist_find(parlist, buffer);
        cpl_parameter_set_int(par, 1);
        sprintf(buffer, "detmon.%s.urx%d", recipe, i);
        par = cpl_parameterlist_find(parlist, buffer);
        cpl_parameter_set_int(par, 2);
        sprintf(buffer, "detmon.%s.ury%d", recipe, i);
        par = cpl_parameterlist_find(parlist, buffer);
        cpl_parameter_set_int(par, 2);
    }

    return cpl_error_get_code();
}

static cpl_error_code detmon_lg_mr_phu_qc_reg_kw(
        cpl_propertylist** mplist[],
        int idx,
        cpl_size next,
        cpl_size nregions,
        const char * kw_segment,
        cpl_type kw_type
        ) {

    char * kw_input = NULL;
    char* kw_output_avg = NULL ;
    char* kw_output_rms = NULL;
    cpl_array* kw_vals = cpl_array_new(next, kw_type);

    // Need to ensure the PHU propertylist actually exists - create one
    // if it's NULL
    if (mplist[idx][0] == NULL) {
        mplist[idx][0] = cpl_propertylist_new();
    }

    for (int i=0; i < nregions; i++) {
        kw_input = cpl_sprintf("ESO QC REG%d %s", i+1, kw_segment);
        cpl_msg_debug(cpl_func, "Look for header kw %s", kw_input);
        kw_output_avg = cpl_sprintf("ESO QC REG%d %s AVG", i+1, kw_segment);
        kw_output_rms = cpl_sprintf("ESO QC REG%d %s RMS", i+1, kw_segment);
        cpl_array_fill_window_invalid(kw_vals, 0, next);

        for (cpl_size e = 1 ; e < next + 1 ; ++e) {
            if (cpl_propertylist_has(mplist[idx][e], kw_input)) {
                cpl_msg_debug(
                        cpl_func,
                        "            Found value %s in ext %" CPL_SIZE_FORMAT "",
                        kw_input, e);
                if (kw_type == CPL_TYPE_DOUBLE) {
                    cpl_array_set_double(kw_vals, e-1,
                                         cpl_propertylist_get_double(
                                                 mplist[idx][e], kw_input));
                } else if (kw_type == CPL_TYPE_INT) {
                    cpl_array_set_int(kw_vals, e-1,
                                      cpl_propertylist_get_int(
                                              mplist[idx][e], kw_input));
                }
            }
        }

        cpl_propertylist_append_double(mplist[idx][0],
                                       kw_output_avg,
                                       cpl_array_get_mean(kw_vals));
        cpl_msg_debug(cpl_func, "        Set mean");
        cpl_propertylist_set_comment(
                mplist[idx][0],
                kw_output_avg,
                cpl_sprintf("Avg of QC REG%d %s over exts",
                            i+1, kw_segment));
        cpl_msg_debug(cpl_func, "        Set mean comment");
        cpl_propertylist_append_double(mplist[idx][0],
                                       kw_output_rms,
                                       cpl_array_get_stdev(kw_vals));
        cpl_msg_debug(cpl_func, "        Set RMS");
        cpl_propertylist_set_comment(
                mplist[idx][0],
                kw_output_rms,
                cpl_sprintf("RMS of QC REG%d %s over exts",
                            i+1, kw_segment));
        if (cpl_error_get_code() != CPL_ERROR_NONE) {
            cpl_msg_error(__func__, "Error found in detmon_lg_mr_phu_qc_reg_kw!");
            cpl_msg_error(__func__, "In %s, L%u", cpl_error_get_file(),
                          cpl_error_get_line());
            cpl_msg_error(__func__, "%s", cpl_error_get_message());
            cpl_msg_error(__func__, "Values ingested:");
            cpl_array_dump(kw_vals, 0, next, stdout);
            cpl_msg_error(__func__, "There were %d extensions", next);
            cpl_msg_error(__func__, "There were %d regions detected", nregions);
        }
    }

    cpl_free(kw_input);
    cpl_free(kw_output_rms);
    cpl_free(kw_output_avg);
    cpl_array_delete(kw_vals);

    return cpl_error_get_code();
}

/*---------------------------------------------------------------------------*/
/*
  @brief    Interpret the command line options and execute the data processing
  @param    frameset    the frames list
  @param    parlist_    the parameters list
  @param    recipe      recipe name
  @param    is_nir      switch to know if opt or nir
  @return   0 if everything is ok
 */
/*---------------------------------------------------------------------------*/

int detmon_lg_mr(cpl_frameset            * frameset,
                 const cpl_parameterlist * parlist_,
                 const char * recipe,
                 cpl_boolean is_nir)
{
    const char * tag_on=NULL;
    const char * tag_off=NULL;
    IRPLIB_DIAG_PRAGMA_PUSH_IGN(-Wcast-qual);
    cpl_parameterlist * parlist = (cpl_parameterlist *)parlist_;
    IRPLIB_DIAG_PRAGMA_POP;
    cpl_propertylist ** mplist[DETMON_MAX_IDX] = {NULL};
    cpl_imagelist ** mimg[DETMON_MAX_IDX] = {NULL};
    cpl_table ** mtab[DETMON_MAX_IDX] = {NULL};
    cpl_size next[DETMON_MAX_IDX] = {0};
    char * outname[DETMON_MAX_IDX] = {NULL};
    char * outtag[DETMON_MAX_IDX] = {NULL};
    cpl_size nregions = 0;

    cpl_frameset * frameset_copy = NULL;

    cpl_propertylist * lintbl =
        detmon_fill_prolist("DET_LIN_INFO", "REDUCED", "TECH", CPL_FALSE);

    cpl_propertylist * gaintbl =
        detmon_fill_prolist("GAIN_INFO",    "REDUCED", "TECH", CPL_FALSE);

    cpl_propertylist * coeffscube =
        detmon_fill_prolist("COEFFS_CUBE",  "REDUCED", "TECH", CPL_FALSE);

    cpl_propertylist * bpm =
        detmon_fill_prolist("BP_MAP_NL",    "REDUCED", "TECH", CPL_FALSE);

    cpl_propertylist * corr =
        detmon_fill_prolist("AUTOCORR",     "REDUCED", "TECH", CPL_FALSE);

    cpl_propertylist * diff_flat =
        detmon_fill_prolist("DIFF_FLAT",    "REDUCED", "TECH", CPL_FALSE);
    const char * regions_file = NULL;
    const char * regions_str = NULL;
    cpl_error_code error = CPL_ERROR_NONE;
    cpl_size * llx = NULL, * lly = NULL, * urx = NULL, * ury = NULL;
    const cpl_parameter * p =
        get_parameter_const(parlist, recipe, "regions-file");
    if (p) {
        regions_file = cpl_parameter_get_string(p);
        if (strlen(regions_file) == 0)
            regions_file = NULL;
    }

    p = get_parameter_const(parlist, recipe, "regions");
    if (p) {
        regions_str = cpl_parameter_get_string(p);
        if (strlen(regions_str) == 0)
            regions_str = NULL;
        else {
            if (regions_file) {
                return cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT,
                                             "Please only provide either "
                                             "--regions or "
                                             "--regions-file, not both");
            }
        }
    }

    p = get_parameter(parlist, recipe, "intermediate");
    if (p) {
        error_if(cpl_parameter_get_bool(p), CPL_ERROR_UNSUPPORTED_MODE,
                 "--intermediate not supported by multi region recipe");
    }

    skip_if (detmon_lg_set_tag(frameset, &tag_on, &tag_off));
    skip_if (detmon_lg_dfs_set_groups(frameset, tag_on, tag_off));

    if (regions_file) {
        read_regions_file(regions_file, &llx, &lly, &urx, &ury,
                          &nregions);
    }
    else if (regions_str) {
        read_regions_str(regions_str, &llx, &lly, &urx, &ury,
                         &nregions);
    }
    else {
        return cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT,
                                     "Empty value in --regions or "
                                     "--regions-file");
    }
    skip_if(0);

    skip_if(check_regions(llx, lly, urx, ury, nregions));

    if (!is_nir) {
        /* contamination not supported */
        skip_if(disable_contamination(parlist, recipe));
    }

    /* additionally iterating on extensions would improve IO caching and
     * reduce memory usage for pix2pix but requires checks if detmon really
     * gives the same result in this case */
    for (int i = 0; i < nregions; i++) {
        cpl_size nlabels;
        cpl_size * selection;
        cpl_parameter * par;
        par = get_parameter(parlist, recipe, "llx");
        cpl_parameter_set_int(par, llx[i]);
        par = get_parameter(parlist, recipe, "lly");
        cpl_parameter_set_int(par, lly[i]);
        par = get_parameter(parlist, recipe, "urx");
        cpl_parameter_set_int(par, urx[i]);
        par = get_parameter(parlist, recipe, "ury");
        cpl_parameter_set_int(par, ury[i]);
        cpl_msg_info(cpl_func, "working on region %d/%d: %lld %lld, "
                     "%lld %lld", i + 1, (int)nregions,
                     llx[i], lly[i], urx[i], ury[i]);

        /* copy and throw away so we don't accumulate products
         * for each call */
        cpl_frameset_delete(frameset_copy);
        frameset_copy = cpl_frameset_duplicate(frameset);
        error = detmon_lg(frameset_copy,
                          parlist,
                          tag_on,
                          tag_off,
                          recipe,
                          PACKAGE_TARNAME,
                          REGEXP,
                          lintbl, gaintbl, coeffscube,
                          bpm, corr, diff_flat,
                          PACKAGE "/" PACKAGE_VERSION,
                          NULL, NULL, is_nir);
        skip_if(error);

        selection = cpl_frameset_labelise(frameset_copy, compare_tags,
                                          &nlabels);

        for (cpl_size j = 0; j < nlabels; j++) {
            cpl_frameset * a = cpl_frameset_extract(frameset_copy, selection, j);
            cpl_frame * frm = cpl_frameset_get_position(a, 0);
            const char * fn = cpl_frame_get_filename(frm);
            int idx;

            if (strcmp(cpl_frame_get_tag(frm), "DET_LIN_INFO") == 0)
                idx = DETMON_LIN_INFO_IDX;
            else if (strcmp(cpl_frame_get_tag(frm), "GAIN_INFO") == 0)
                idx = DETMON_GAIN_INFO_IDX;
            else if (strcmp(cpl_frame_get_tag(frm), "BP_MAP_NL") == 0)
                idx = DETMON_BPM_IDX;
            else if (strcmp(cpl_frame_get_tag(frm), "COEFFS_CUBE") == 0)
                idx = DETMON_COEFF_IDX;
            else {
                cpl_frameset_delete(a);
                continue;
            }

            /* prepare storage */
            if (mplist[idx] == NULL) {
                next[idx] = cpl_frame_get_nextensions(frm);
                outname[idx] = cpl_strdup(cpl_frame_get_filename(frm));
                outtag[idx] = cpl_strdup(cpl_frame_get_tag(frm));
                mplist[idx] = cpl_calloc(sizeof(cpl_propertylist*), next[idx] + 1);
                mimg[idx] = cpl_calloc(sizeof(cpl_imagelist*), next[idx] + 1);
                mtab[idx] = cpl_calloc(sizeof(cpl_table*), next[idx] + 1);
            }

            if (idx == DETMON_LIN_INFO_IDX) {
                const char * rmregexp = "ESO QC CONTAM.*";
                merge_lingain(mplist[idx], mtab[idx], fn, next[idx],
                              rmregexp, i + 1);
            }
            else if (idx == DETMON_GAIN_INFO_IDX) {
                merge_lingain(mplist[idx], mtab[idx], fn, next[idx],
                              NULL, i + 1);
            }
            else if (idx == DETMON_BPM_IDX) {
                merge_bpmcoeff(mplist[idx], mimg[idx], fn, next[idx],
                              NULL, i + 1);
            }
            else if (idx == DETMON_COEFF_IDX) {
                merge_bpmcoeff(mplist[idx], mimg[idx], fn, next[idx],
                              NULL, i + 1);
            }
            skip_if(0);

            for (int e = 0; e < next[idx] + 1; e++) {
                char buffer[300];
                cpl_propertylist * copy_plist;
                cpl_size len;
                if (!mplist[idx][e])
                    continue;

                /* don't add region if extension has no QC parameters */
                copy_plist = cpl_propertylist_duplicate(mplist[idx][e]);
                cpl_propertylist_erase_regexp(copy_plist, "ESO QC.*", 1);
                len = cpl_propertylist_get_size(copy_plist);
                cpl_propertylist_delete(copy_plist);
                if (len == 0) {
                    continue;
                }

                sprintf(buffer, "ESO QC REG%d LLX", i + 1);
                cpl_propertylist_append_int(mplist[idx][e], buffer, llx[i]);
                cpl_propertylist_set_comment(mplist[idx][e], buffer,
                                             "Lower left X of region");
                sprintf(buffer, "ESO QC REG%d LLY", i + 1);
                cpl_propertylist_append_int(mplist[idx][e], buffer, lly[i]);
                cpl_propertylist_set_comment(mplist[idx][e], buffer,
                                             "Lower left Y of region");
                sprintf(buffer, "ESO QC REG%d URX", i + 1);
                cpl_propertylist_append_int(mplist[idx][e], buffer, urx[i]);
                cpl_propertylist_set_comment(mplist[idx][e], buffer,
                                             "Upper right X of region");
                sprintf(buffer, "ESO QC REG%d URY", i + 1);
                cpl_propertylist_append_int(mplist[idx][e], buffer, ury[i]);
                cpl_propertylist_set_comment(mplist[idx][e], buffer,
                                             "Upper right Y of region");
            }

            cpl_frameset_delete(a);
            skip_if(0);
        }
        cpl_free(selection);
    }
    cpl_msg_debug(__func__, "Completed looping over regions");

    {
        cpl_frameset * usedframes = cpl_frameset_new();
        const cpl_size nframes    = cpl_frameset_get_size(frameset);

        cpl_frameset * rawframes = cpl_frameset_new();

        for (cpl_size idx = 0; idx < nframes; idx++) {
            const cpl_frame * frm = cpl_frameset_get_position_const(frameset,
                                                                    idx);
            if (cpl_frame_get_group(frm) == CPL_FRAME_GROUP_RAW ||
                cpl_frame_get_group(frm) == CPL_FRAME_GROUP_CALIB) {
                cpl_frameset_insert(usedframes, cpl_frame_duplicate(frm));
            }
            if (cpl_frame_get_group(frm) == CPL_FRAME_GROUP_RAW) {
                cpl_frameset_insert(rawframes, cpl_frame_duplicate(frm));
            }
        }

        cpl_array* reg_values = NULL;
        char * var_part = NULL;

        for (int idx = 0; idx < DETMON_MAX_IDX; idx++) {
            cpl_msg_debug(__func__, "Doing idx loop %d of %d", idx,
                         DETMON_MAX_IDX-1);
            if (!mplist[idx])
                continue;

            if ((idx == DETMON_LIN_INFO_IDX || idx == DETMON_GAIN_INFO_IDX)) {

                // PIPE-10123 - Compute additional parameters here for
                // these two tables
                double exptime = 0.0 ;
                reg_values = cpl_array_new(nregions,
                                           CPL_TYPE_DOUBLE);

                if (!is_nir && next[idx] > 0) {
                    cpl_msg_debug(cpl_func, "Doing PIPE-10123 lin/gain kw for %s",
                                  recipe);

                    // Compute EXPTIME MAX from RAWFRAMES
                    for (cpl_size r=0;
                    r < cpl_frameset_get_size(rawframes);
                    ++r) {
                        const cpl_frame * frm = cpl_frameset_get_position_const(
                                rawframes, r
                                );
                        const char * rfn = cpl_frame_get_filename(frm);
                        const cpl_propertylist * rpl = cpl_propertylist_load(
                                rfn, 0
                                );
                        if (rpl != NULL &&
                        cpl_propertylist_has(rpl, "EXPTIME")) {
                            if (cpl_propertylist_get_double(rpl, "EXPTIME") >
                            exptime) {
                                exptime = cpl_propertylist_get_double(
                                        rpl, "EXPTIME");
                            }
                        }
                    }
                    cpl_propertylist_append_double(mplist[idx][0],
                                                   "ESO QC EXPTIME MAX",
                                                   exptime);
                    cpl_propertylist_set_comment(
                            mplist[idx][0],
                            "ESO QC EXPTIME MAX",
                            "Max. EXPTIME from input raws");

                    if (idx == DETMON_GAIN_INFO_IDX) {
                        var_part = "GAIN";
                        detmon_lg_mr_phu_qc_reg_kw(mplist, idx, next[idx],
                                                   nregions, "CONAD",
                                                   CPL_TYPE_DOUBLE);
                        detmon_lg_mr_phu_qc_reg_kw(mplist, idx, next[idx],
                                                   nregions, "GAIN",
                                                   CPL_TYPE_DOUBLE);
                    } else {
                        var_part = "LIN EFF";
                        detmon_lg_mr_phu_qc_reg_kw(mplist, idx, next[idx],
                                                   nregions, "GAIN ERR",
                                                   CPL_TYPE_DOUBLE);
                        detmon_lg_mr_phu_qc_reg_kw(mplist, idx, next[idx],
                                                   nregions, "FPN",
                                                   CPL_TYPE_DOUBLE);
                        detmon_lg_mr_phu_qc_reg_kw(mplist, idx, next[idx],
                                                   nregions, "COUNTS MIN",
                                                   CPL_TYPE_DOUBLE);
                        detmon_lg_mr_phu_qc_reg_kw(mplist, idx, next[idx],
                                                   nregions, "COUNTS MAX",
                                                   CPL_TYPE_DOUBLE);
                        detmon_lg_mr_phu_qc_reg_kw(mplist, idx, next[idx],
                                                   nregions, "LIN EFF",
                                                   CPL_TYPE_DOUBLE);
                        detmon_lg_mr_phu_qc_reg_kw(mplist, idx, next[idx],
                                                   nregions, "LIN EFF FLUX",
                                                   CPL_TYPE_DOUBLE);
                        detmon_lg_mr_phu_qc_reg_kw(mplist, idx, next[idx],
                                                   nregions, "LLX",
                                                   CPL_TYPE_INT);
                        detmon_lg_mr_phu_qc_reg_kw(mplist, idx, next[idx],
                                                   nregions, "LLY",
                                                   CPL_TYPE_INT);
                        detmon_lg_mr_phu_qc_reg_kw(mplist, idx, next[idx],
                                                   nregions, "URX",
                                                   CPL_TYPE_INT);
                        detmon_lg_mr_phu_qc_reg_kw(mplist, idx, next[idx],
                                                   nregions, "URY",
                                                   CPL_TYPE_INT);
                    }

                    for (cpl_size e = 1; e < next[idx] + 1; e++) {

                        for (int i=0 ; i < nregions ; i++) {
                            if (cpl_propertylist_has(mplist[idx][e],
                                                     cpl_sprintf("ESO QC REG%d %s",
                                                                 i+1, var_part))) {
                                cpl_array_set_double(
                                        reg_values, i,
                                        cpl_propertylist_get_double(mplist[idx][e],
                                                                    cpl_sprintf("ESO QC REG%d %s",
                                                                                i+1, var_part)
                                                                                ));
                            }

                        }

                        cpl_propertylist_append_double(mplist[idx][e],
                                                       cpl_sprintf("ESO QC %s AVG",
                                                                   var_part),
                                                                   cpl_array_get_mean(
                                                                           reg_values));
                        cpl_propertylist_set_comment(mplist[idx][e],
                                                     cpl_sprintf("ESO QC %s AVG",
                                                                 var_part),
                                                                 cpl_sprintf(
                                                                         "Avg of QC.REGn.%s",
                                                                         var_part));
                        cpl_propertylist_append_double(mplist[idx][e],
                                                       cpl_sprintf("ESO QC %s RMS",
                                                                   var_part),
                                                                   cpl_array_get_stdev(
                                                                           reg_values));
                        cpl_propertylist_set_comment(mplist[idx][e],
                                                     cpl_sprintf("ESO QC %s RMS",
                                                                 var_part),
                                                                 cpl_sprintf(
                                                                         "RMS of QC.REGn.%s",
                                                                         var_part));
                        cpl_msg_debug(__func__, "Wrote ESO QC RMS");
                    }

                }
                cpl_msg_debug(__func__, "Doing a GAIN/LIN save (var_part=%s)",
                              var_part);
                irplib_dfs_save_propertylist(frameset, parlist, usedframes,
                                             recipe, outtag[idx],
                                             mplist[idx][0], NULL, PACKAGE "/"
                                             PACKAGE_VERSION, outname[idx]);

                skip_if(0);
                for (cpl_size e = 1; e < next[idx] + 1; e++) {
                    if (mtab[idx][e]) {
                        cpl_table_save(mtab[idx][e], NULL, mplist[idx][e],
                                       outname[idx], CPL_IO_EXTEND);
                    }
                    else {
                        cpl_propertylist_save(mplist[idx][e], outname[idx],
                                              CPL_IO_EXTEND);
                    }
                    skip_if(0);
                }
            }
            // No need to alter _mr BPM after all for PIPE-10123
            else if (idx == DETMON_BPM_IDX) {
                cpl_msg_debug(__func__, "Doing a BPM save");
                if (!mimg[idx][0]) {
                    irplib_dfs_save_propertylist(frameset, parlist, usedframes,
                                                 recipe, outtag[idx],
                                                 mplist[idx][0], NULL, PACKAGE
                                                 "/" PACKAGE_VERSION,
                                                 outname[idx]);
                }
                else {
                    irplib_dfs_save_image(frameset, parlist, usedframes,
                                          cpl_imagelist_get(mimg[idx][0], 0),
                                          CPL_TYPE_UNSPECIFIED, recipe,
                                          outtag[idx], mplist[idx][0], NULL,
                                          PACKAGE "/" PACKAGE_VERSION,
                                          outname[idx]);
                    cpl_imagelist_delete(mimg[idx][0]);
                    mimg[idx][0] = NULL;
                }
                skip_if(0);
                for (cpl_size e = 1; e < next[idx] + 1; e++) {
                    if (mimg[idx][e]) {
                        cpl_image * img = cpl_imagelist_get(mimg[idx][e], 0);
                        cpl_image_save(img, outname[idx], CPL_TYPE_UNSPECIFIED,
                                       mplist[idx][e],
                                       CPL_IO_EXTEND);
                        cpl_imagelist_delete(mimg[idx][e]);
                        mimg[idx][e] = NULL;
                    }
                    else {
                        cpl_propertylist_save(mplist[idx][e], outname[idx],
                                              CPL_IO_EXTEND);
                    }
                    skip_if(0);
                }
            }
            else if (idx == DETMON_COEFF_IDX) {
                // PIPE-10123 Add the additional QC keywords
                if (!is_nir && next[idx] > 0) {
                    cpl_msg_debug(cpl_func, "Doing PIPE-10123 COEFF kw for %s",
                                  recipe);
                    for (int j=0; j < 4 ; ++j ) {
                        detmon_lg_mr_phu_qc_reg_kw(mplist, idx, next[idx],
                                                   nregions,
                                                   cpl_sprintf("LIN COEF%d", j),
                                                   CPL_TYPE_DOUBLE);
                    }
                    detmon_lg_mr_phu_qc_reg_kw(mplist, idx, next[idx],
                                               nregions, "ERRFIT",
                                               CPL_TYPE_DOUBLE);
                }

                cpl_msg_debug(__func__, "Doing a COEFFS save");
                if (!mimg[idx][0]) {
                    irplib_dfs_save_propertylist(frameset, parlist, usedframes,
                                                 recipe, outtag[idx],
                                                 mplist[idx][0], NULL, PACKAGE
                                                 "/" PACKAGE_VERSION,
                                                 outname[idx]);
                }
                else  {
                    irplib_dfs_save_imagelist(frameset, parlist, usedframes,
                                              mimg[idx][0],
                                              CPL_TYPE_UNSPECIFIED,
                                              recipe, outtag[idx],
                                              mplist[idx][0], NULL, PACKAGE "/"
                                              PACKAGE_VERSION, outname[idx]);
                    cpl_imagelist_delete(mimg[idx][0]);
                    mimg[idx][0] = NULL;
                }
                skip_if(0);
                for (cpl_size e = 1; e < next[idx] + 1; e++) {
                    if (mimg[idx][e]) {
                        cpl_imagelist_save(mimg[idx][e], outname[idx],
                                           CPL_TYPE_UNSPECIFIED,
                                           mplist[idx][e], CPL_IO_EXTEND);
                        cpl_imagelist_delete(mimg[idx][e]);
                        mimg[idx][e] = NULL;
                    }
                    else {
                        cpl_propertylist_save(mplist[idx][e], outname[idx],
                                              CPL_IO_EXTEND);
                    }
                    skip_if(0);
                }
            }
        }
        // PIPE-10123 cleanup
        cpl_array_delete(reg_values);

        cpl_frameset_delete(usedframes);
        cpl_frameset_delete(rawframes);
        skip_if(0);
    }

    end_skip;
    cpl_frameset_delete(frameset_copy);
    cpl_propertylist_delete(lintbl);
    cpl_propertylist_delete(gaintbl);
    cpl_propertylist_delete(coeffscube);
    cpl_propertylist_delete(bpm);
    cpl_propertylist_delete(corr);
    cpl_propertylist_delete(diff_flat);
    for (int i = 0; i < DETMON_MAX_IDX; i++) {
        for (int e = 0; e < next[i] + 1; e++) {
            if (mplist[i])
                cpl_propertylist_delete(mplist[i][e]);
            if (mimg[i])
                cpl_imagelist_delete(mimg[i][e]);
            if (mtab[i])
                cpl_table_delete(mtab[i][e]);
        }
        cpl_free(mplist[i]);
        cpl_free(mimg[i]);
        cpl_free(mtab[i]);
        cpl_free(outname[i]);
        cpl_free(outtag[i]);
    }
    cpl_free(llx);
    cpl_free(lly);
    cpl_free(urx);
    cpl_free(ury);

    return cpl_error_get_code();
}
/**@}*/
