/* $Id: detmon.c,v 1.11 2013-07-19 12:00:24 jtaylor Exp $
 *
 * This file is part of the irplib package
 * Copyright (C) 2002, 2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02111-1307 USA
 */

/*
 * $Author: jtaylor $
 * $Date: 2013-07-19 12:00:24 $
 * $Revision: 1.11 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*---------------------------------------------------------------------------
                                  Includes
 ---------------------------------------------------------------------------*/

/* Must include first to ensure declaration of complex CPL functions */
#include <complex.h>

#include "detmon_pernoise.h"
#include "detmon.h"

#include "irplib_ksigma_clip.h"
#include "irplib_hist.h"
#include "irplib_utils.h"

#include <math.h>
#include <string.h>
#include <assert.h>
#include <float.h>

/*----------------------------------------------------------------------------
                             Defines
 ----------------------------------------------------------------------------*/

/* Computes the square of an euclidean distance bet. 2 points */
#define pdist(x1,y1,x2,y2) (((x1-x2)*(x1-x2))+((y1-y2)*(y1-y2)))

#define cpl_drand() ((double)rand()/(double)RAND_MAX)
/**@{*/
/*--------------------------------------------------------------------------*/

/*
 * @defgroup detmon        Detector monitoring functions
 */

/*--------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
                                  Defines
 ---------------------------------------------------------------------------*/

enum pixeltypes
{
    HOT = 0,
    DEAD = 1,
    NOISY = 2
};

enum stackingtypes
{
    MINMAX = 0,
    MEAN = 1,
    MEDIAN = 2,
    KSIGMA = 3
};

enum readouts
{
    HORIZONTAL = 1,
    VERTICAL = 2
};



static struct
{
    int         mode;
    cpl_boolean direction;
    double      speed;
    int         llx;
    int         lly;
    int         urx;
    int         ury;
    int         kappa;
    int         exts;
    int         nb_extensions;
} detmon_pernoise_config;


#define NIR TRUE
#define OPT FALSE

/*---------------------------------------------------------------------------
  Private function prototypes
  ---------------------------------------------------------------------------*/

/* Functions for Periodic Noise Characterisation, detmon_pernoise() */

int
detmon_pernoise_dfs_set_groups(cpl_frameset *,
                               const char *);

static cpl_error_code
detmon_pernoise_retrieve_parlist(const char *,
                                 const char *,
                                 const cpl_parameterlist *);

static cpl_error_code
detmon_pernoise_qc(cpl_propertylist *,
                   cpl_table        *,
                   int);

static cpl_error_code
detmon_pernoise_save(const cpl_parameterlist *,
                     cpl_frameset *,
                     const char *,
                     const char *,
                     const char *,
                     const char *,
                     cpl_table **,
                     cpl_propertylist **,
                     const int,
                     const int,
                     const cpl_frameset *);

cpl_error_code
detmon_pernoise_rm_bg(cpl_image *,
                      int,
                      int);

/*--------------------------------------------------------------------------*/

/*
 * @brief  Reduce periodic noise
 * @param  image         Input image
 *
 * @return 1-d image, representing the distribution of noise. NULL if failed.
 */

/*--------------------------------------------------------------------------*/


#define HORIZONTAL TRUE

cpl_table              *
detmon_pernoise_reduce(cpl_image * image)
{
    int                     nsamples, nffts;
    int                     i, j;

    int status;
    float * hanning = 0;
    float * power = 0;
    cpl_image * power_im = 0;
    cpl_image * output = 0;
    cpl_image * pos_spec = 0;
    cpl_table * table = 0;
    cpl_image* fourier_im = 0;
    double freq;
    cpl_error_code error = CPL_ERROR_NONE;
    cpl_image * sub_image = 0;
    int nffts_old;


    if(detmon_pernoise_config.direction == HORIZONTAL) {
        error = cpl_image_flip(image, 1);
        cpl_ensure(!error, error, NULL);
    }

    nsamples = cpl_image_get_size_x(image);
    nffts = cpl_image_get_size_y(image);


    /* Rewrite the previous lines with a better style (: ...) */

    /*
     * 1. preprocessing task:
     * Estimate the background fitting the image to a
     * 2-D polynomial and substract it from the image.
     */

    error = detmon_pernoise_rm_bg(image, nsamples, nffts);
    cpl_ensure(!error, error, NULL);

    sub_image = cpl_image_extract(image, nsamples/8 + 1, nffts/8+1,
                                  nsamples*7/8, nffts*7/8);
    nffts_old = nffts;
    nsamples = cpl_image_get_size_x(sub_image);
    nffts = cpl_image_get_size_y(sub_image);

    /*
     * 2. preprocessing task:
     * Remove the effect of hot and dark pixels, replacing their values by
     * the average value of the neighbouring pixels.
     */

    /*
     * 3. preprocessing task:
     * Apply a Hanning vector
     */

    hanning = cpl_malloc(sizeof(float) * nsamples);

    for(i = 0; i < nsamples; i++) {
        *(hanning + i) = 0.5 - 0.5 * cos(2 * CPL_MATH_PI * (float) i / nsamples);
        for(j = 0; j < nffts; j++) {
            double                  value =
                cpl_image_get(sub_image, i + 1, j + 1, &status);
            error = cpl_image_set(sub_image, i + 1, j + 1, (*(hanning + i)) * value);
        }
    }

    cpl_free(hanning);
    if (error != CPL_ERROR_NONE)
    {
        goto cleanup;
    }

    power = (float *) cpl_calloc(sizeof(float), nsamples * nffts);


    fourier_im = cpl_image_new(nsamples,nffts,  CPL_TYPE_FLOAT_COMPLEX);
    error = cpl_fft_image(fourier_im, sub_image, CPL_FFT_FORWARD);

    for(i = 1; i <= nffts; i++) {
        for(j = 1; j <= nsamples; j++) {
            int rej = 0;
            double complex cvalue = cpl_image_get_complex(fourier_im,j, i, &rej );
            double value = cabs(cvalue);
            /*
             *(power + j + i * nsamples) = *((float complex *)fourier + j + i * nsamples) *
             conjf(*((float complex *)fourier +j + i * nsamples));
             */
            cpl_image_set(power_im, j, i, value);
        }
        /*  Is it necessary to divide here by 2 * pi? */
    }
    cpl_image_delete(fourier_im);

    output   = cpl_image_collapse_create(power_im, 0);
    pos_spec = cpl_image_extract(output, 1, 1, nsamples/2, 1);

    cpl_image_delete(power_im);
    cpl_free(power);

    cpl_image_delete(output);

    table = cpl_table_new(nsamples/2);
    cpl_table_new_column(table, "FREQ", CPL_TYPE_DOUBLE);
    cpl_table_new_column(table, "POW", CPL_TYPE_DOUBLE);

    freq = detmon_pernoise_config.speed*1000/nffts_old;

    for(i = 0; i < nsamples/2; i++) {
        error = cpl_table_set(table, "FREQ", i, freq/(nsamples/2)*i);
        error = cpl_table_set(table, "POW", i, cpl_image_get(pos_spec, i+1, 1, &status));
    }

    for(i= 0; i < 5; i++) {
        error = cpl_table_set(table, "POW", i, 0.0);
    }


cleanup:
    cpl_image_delete(pos_spec);

    cpl_image_delete(sub_image);
    if (error != CPL_ERROR_NONE)
    {
        cpl_table_delete(table);
        table = 0;
    }
    return table;
}
#undef HORIZONTAL



/*---------------------------------------------------------------------------*/

/*
 * @brief  Measure persistence noise
 * @param  frameset Input set of frames
 * @param  parlist  Input parameterlist
 * @param  tag      Input tag
 * @param  recipe_name input recipe name
 * @param  pipeline_name input pipeline id
 * @param  ptrocatg_tbl   input table with pro catgs
 * @param  package        package id
 * @param  comapre        comparison function
 * @return CPL_ERROR_NONE on success.
 */

/*---------------------------------------------------------------------------*/
cpl_error_code
detmon_pernoise(cpl_frameset      * frameset,
                const cpl_parameterlist * parlist,
                const char        * tag,
                const char        * recipe_name,
                const char        * pipeline_name,
                const char        * procatg_tbl,
                const char        * package,
                int              (*compare)(const cpl_frame *,
                                            const cpl_frame *))
{
    cpl_size                     nsets;

    int                     i;
    cpl_error_code          error;

    if(detmon_pernoise_dfs_set_groups(frameset, tag)) {
        cpl_msg_error(cpl_func, "Cannot identify RAW and CALIB frames");
    }

    /*
     * This function reads all inputs parameters from parlist
     * and stores them in a global variable detmon_ronbias_config.
     * Similar to detmon_lg_retrieve_parlist(). See detmon.c
     */
    error = detmon_pernoise_retrieve_parlist(pipeline_name,
                                             recipe_name, parlist);
    cpl_ensure_code(!error, error);

    /* Labelise all input frames */
    if(compare == NULL)
        nsets = 1;
    else {
        cpl_msg_info(cpl_func, "Identify the different settings");
        cpl_size* selection = cpl_frameset_labelise(frameset, compare, &nsets);
        if(selection == NULL)
            cpl_msg_error(cpl_func, "Cannot labelise input frames");
    }

    detmon_pernoise_config.nb_extensions = 1;
    if(detmon_pernoise_config.exts < 0) {
        const cpl_frame        *cur_frame =
            cpl_frameset_get_position_const(frameset, 0);
        /* Get the nb of extensions */
        detmon_pernoise_config.nb_extensions =
            cpl_frame_get_nextensions(cur_frame);
    }

    /* Extract settings and reduce each of them */
    for(i = 0; i < nsets; i++)
    {
        int j;
        cpl_table ** freq_table;
        cpl_propertylist ** qclist =
            (cpl_propertylist **)
            cpl_malloc(detmon_pernoise_config.nb_extensions *
                       sizeof(cpl_propertylist *));

        cpl_imagelist ** raws = (cpl_imagelist **) cpl_malloc(detmon_pernoise_config.nb_extensions * sizeof(cpl_imagelist *));
        cpl_image ** input = (cpl_image **) cpl_malloc(detmon_pernoise_config.nb_extensions * sizeof(cpl_image *));

        /* Initialise memory for products */
        if(detmon_pernoise_config.mode == 1)
        {
            freq_table =
                (cpl_table **) cpl_malloc(detmon_pernoise_config.nb_extensions *
                                          4 * sizeof(cpl_table *));
        } else
        {
            freq_table =
                (cpl_table **) cpl_malloc(detmon_pernoise_config.nb_extensions *
                                          sizeof(cpl_table *));
        }

        if(detmon_pernoise_config.exts >= 0)
        {
            *raws =
                cpl_imagelist_load_frameset(frameset, CPL_TYPE_FLOAT, 1,
                                            detmon_pernoise_config.exts);
            *input = cpl_image_subtract_create(cpl_imagelist_get(*raws,0),
                                               cpl_imagelist_get(*raws,1));
        } else
        {
            cpl_imagelist          *raws_all_exts =
                cpl_imagelist_load_frameset(frameset, CPL_TYPE_FLOAT, 1,
                                            -1);
            for(j = 0; j < detmon_pernoise_config.nb_extensions; j++)
            {
                int nframes = cpl_frameset_get_size(frameset);
                int                     k;
                for(k = 0; k < nframes; k++)
                {
                    cpl_image              *image =
                        cpl_imagelist_unset(raws_all_exts,
                                            (detmon_pernoise_config.
                                             nb_extensions - 1 - j) * k);
                    cpl_imagelist_set(raws[j], image, k);
                }
                input[j] =
                    cpl_image_subtract_create(cpl_imagelist_get(raws[j],0),
                                              cpl_imagelist_get(raws[j],1));
            }
        }

        for(j = 0; j < detmon_pernoise_config.nb_extensions; j++) {
            cpl_msg_info(cpl_func, "Starting reduction");
            qclist[j] = cpl_propertylist_new();
            if(detmon_pernoise_config.mode == 1)
            {
                int nx = cpl_image_get_size_x(input[j]);
                int ny = cpl_image_get_size_y(input[j]);
                int k = 0;
                cpl_image* quad[4];

                quad[0] = cpl_image_extract(input[j], 1, 1, nx/2, ny/2);
                quad[1] = cpl_image_extract(input[j], 1, ny/2+1, nx/2, ny);
                quad[2] = cpl_image_extract(input[j], nx/2+1, 1, nx, ny/2);
                quad[3] = cpl_image_extract(input[j], nx/2+1, ny/2+1, nx, ny);

                for (k = 0; k < 4; k++)
                {
                    freq_table[j * 4 + k] = detmon_pernoise_reduce(quad[k]);
                }
                for(k = 0; k < 4; k++)
                {
                    error = detmon_pernoise_qc(qclist[j], freq_table[j + k], k+1);
                    if (error != CPL_ERROR_NONE)
                        break;
                }
                for (k = 0; k < 4; k++)
                {
                    cpl_image_delete(quad[k]);
                }
            } else
            {
                freq_table[j] = detmon_pernoise_reduce(input[j]);
                if(freq_table[j] != NULL)
                {
                    error = detmon_pernoise_qc(qclist[j], freq_table[j], 0);
                }
            }
            if (error != CPL_ERROR_NONE)
            {
                break;
            }
        }
        if (error == CPL_ERROR_NONE)
        {
            error = detmon_pernoise_save(parlist, frameset, recipe_name,
                                         pipeline_name, procatg_tbl,
                                         package, freq_table, qclist, 0,
                                         0, frameset);
        }

        for(j = 0; j < detmon_pernoise_config.nb_extensions; j++)
        {
            cpl_propertylist_delete(qclist[j]);
            cpl_imagelist_delete(raws[j]);
            cpl_image_delete(input[j]);
        }
        cpl_free(qclist);
        cpl_free(raws);
        cpl_free(input);
        if(detmon_pernoise_config.mode == 1)
        {
            for(j= 0; j < detmon_pernoise_config.nb_extensions * 4; j++) {
                cpl_table_delete(freq_table[j]);
            }
        } else {
            for(j= 0; j < detmon_pernoise_config.nb_extensions; j++) {
                cpl_table_delete(freq_table[j]);
            }
        }
        cpl_free(freq_table);
        if (error != CPL_ERROR_NONE)
        {
            break;
        }
    }

    return cpl_error_get_code();
}

/*---------------------------------------------------------------------------*/

/*
 * @brief  classifies input frames matching input tag
 * @param  set        input frameset
 * @param  tag         input tag
 * @return 0 on success, else -1
 */

/*---------------------------------------------------------------------------*/
int
detmon_pernoise_dfs_set_groups(cpl_frameset * set, const char *tag)
{


    int                     nframes;
    int                     i;

    /* Check entries */
    if(set == NULL)
        return -1;

    /* Initialize */
    nframes = cpl_frameset_get_size(set);

    /* Loop on frames */
    for(i = 0; i < nframes; i++) {
        cpl_frame* cur_frame = cpl_frameset_get_position(set, i);
        const char* cur_tag = cpl_frame_get_tag(cur_frame);

        /* RAW frames */
        if(!strcmp(cur_tag, tag))
            cpl_frame_set_group(cur_frame, CPL_FRAME_GROUP_RAW);
        /* CALIB frames */

        /*        else if (!strcmp(tag, IIINSTRUMENT_CALIB_FLAT))
                  cpl_frame_set_group(cur_frame, CPL_FRAME_GROUP_CALIB) ;
                  */
    }
    return 0;
}

/*---------------------------------------------------------------------------*/

/*
 * @brief  Initialize pernoise recipe parameters structure
 * @param  parlist              input parameter list
 * @param  recipe_name          Input recipe name
 * @param  pipeline_name        Input pipeline name
 * @param  mode                 mode to indicate if pix2pix,opt/nir,etc
 * @param  direction            read-out direction
 * @param  speed                read-out speed
 * @param  llx                  lower-left X coordinate
 * @param  lly                  lower-left Y coordinate
 * @param  urx                  upper-right X coordinate
 * @param  ury                  upper-right Y coordinate
 * @param  kappa                kappa value in kappa-sigma clip
 * @param  exts                 extension id

 * @return CPL_ERROR_NONE on success.
 */

/*---------------------------------------------------------------------------*/
    cpl_error_code
detmon_fill_pernoise_params(cpl_parameterlist * parlist,
                            const char *recipe_name,
                            const char *pipeline_name,
                            int mode,
                            const char * direction,
                            double speed,
                            int llx,
                            int lly,
                            int urx,
                            int ury,
                            double kappa,
                            int exts)
{
    detmon_fill_parlist(parlist, recipe_name, pipeline_name, 9,

                        "mode",
                        "Mode",
                        "CPL_TYPE_INT", mode,

                        "direction",
                        "Readout direction",
                        "CPL_TYPE_BOOL", direction,

                        "speed",
                        "Readout speed",
                        "CPL_TYPE_DOUBLE", speed,

                        "llx",
                        "(yet unsupported) x coordinate of the lower-left "
                        "point of the region of interest. If not modified, default value will be 1.",
                        "CPL_TYPE_INT", llx,
                        "lly",
                        "(yet unsupported) y coordinate of the lower-left "
                        "point of the region of interest. If not modified, default value will be 1.",
                        "CPL_TYPE_INT", lly,
                        "urx",
                        "(yet unsupported) x coordinate of the upper-right "
                            "point of the region of interest. If not modified, default value will be X dimension of the input image.",
                        "CPL_TYPE_INT", urx,
                        "ury",
                        "(yet unsupported) y coordinate of the upper-right "
                            "point of the region of interest. If not modified, default value will be Y dimension of the input image.",
                        "CPL_TYPE_INT", ury,

                        "kappa",
                        "Kappa used for determining threshold of bad (hot, cold) pixels",
                        "CPL_TYPE_DOUBLE", kappa,

                        "exts",
                        "Activate the multi-exts option",
                        "CPL_TYPE_INT", exts);

    return 0;
}

/*---------------------------------------------------------------------------*/
/*
 * @brief  initialise pernoise parameter structure
 * @param  parlist              Shift to apply on the x-axis
 * @param  recipe_name          Input recipe id
 * @param  pipeline_name        Input pipeline id
 * @return 0  .
 */

/*---------------------------------------------------------------------------*/
int
detmon_fill_pernoise_params_default(cpl_parameterlist * parlist,
                                    const char *recipe_name,
                                    const char *pipeline_name)
{
    detmon_fill_pernoise_params(parlist, recipe_name, pipeline_name,
                                1,           /* --mode      */
                                "CPL_TRUE",  /* --direction */
                                84.5,        /* --speed     */
                                -1,          /* --llx       */
                                -1,          /* --lly       */
                                -1,          /* --urx       */
                                -1,          /* --ury       */
                                100,          /* --kappa     */
                                0);          /* --exts      */

    return 0;

}

/*---------------------------------------------------------------------------*/
/*
 * @brief  Retrieve pernoise parameter structure values
 * @param  pipeline_name        Input pipeline id
 * @param  recipe_name          Input recipe id
 * @param  parlist              Shift to apply on the x-axis
 * @return 0  .
 */
/*---------------------------------------------------------------------------*/

static cpl_error_code
detmon_pernoise_retrieve_parlist(const char *pipeline_name,
                                 const char *recipe_name,
                                 const cpl_parameterlist * parlist)
{
    char                   *par_name;
    const cpl_parameter    *par;

    /* --mode */
    detmon_pernoise_config.mode =
        detmon_retrieve_par_int("mode", pipeline_name, recipe_name,
                                parlist);

    /* --direction */
    par_name = cpl_sprintf("%s.%s.direction", pipeline_name, recipe_name);
    assert(par_name != NULL);
    par = cpl_parameterlist_find_const(parlist, par_name);
    detmon_pernoise_config.direction = cpl_parameter_get_bool(par);
    cpl_free(par_name);

    /* --speed */
    par_name = cpl_sprintf("%s.%s.speed", pipeline_name, recipe_name);
    assert(par_name != NULL);
    par = cpl_parameterlist_find_const(parlist, par_name);
    detmon_pernoise_config.speed = cpl_parameter_get_double(par);
    cpl_free(par_name);

    /* --llx */
    detmon_pernoise_config.llx =
        detmon_retrieve_par_int("llx", pipeline_name, recipe_name,
                                parlist);

    /* --lly */
    detmon_pernoise_config.lly =
        detmon_retrieve_par_int("lly", pipeline_name, recipe_name,
                                parlist);
    /* --urx */
    detmon_pernoise_config.urx =
        detmon_retrieve_par_int("urx", pipeline_name, recipe_name,
                                parlist);
    /* --ury */
    detmon_pernoise_config.ury =
        detmon_retrieve_par_int("ury", pipeline_name, recipe_name,
                                parlist);
    /* --kappa */
    detmon_pernoise_config.kappa =
        detmon_retrieve_par_double("kappa", pipeline_name, recipe_name,
                                   parlist);

    /* --exts */
    detmon_pernoise_config.exts =
        detmon_retrieve_par_int("exts", pipeline_name, recipe_name,
                                parlist);

    if(cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "Failed to retrieve the input parameters");
        cpl_ensure_code(0, CPL_ERROR_DATA_NOT_FOUND);
    }

    return cpl_error_get_code();
}

/*---------------------------------------------------------------------------*/
/*
 * @brief  Save results of pernoise recipe
 * @param  parlist              input parameter list
 * @param  frameset             input frameset
 * @param  recipe_name          input recipe name
 * @param  pipeline_name        input pipeline name
 * @param  procatg_tbl          input procategory table
 * @param  package              input package id
 * @param  freq_table           computed frequency table
 * @param  qclist               computed QC parameters
 * @param  flag_sets            switch
 * @param  which_sets           specifier
 * @param  usedframes           used frames

 * @return CPL_ERROR_NONE on success.
 */
/*---------------------------------------------------------------------------*/
static                  cpl_error_code
detmon_pernoise_save(const cpl_parameterlist * parlist,
                     cpl_frameset * frameset,
                     const char *recipe_name,
                     const char *pipeline_name,
                     const char *procatg_tbl,
                     const char *package,
                     cpl_table ** freq_table,
                     cpl_propertylist ** qclist,
                     const int flag_sets,
                     const int which_set,
                     const cpl_frameset * usedframes)
{

    cpl_frame              *ref_frame;
    cpl_propertylist       *plist;
    char                   *name_o = NULL; /* Avoid (false) uninit warning */
    int                     i;
    cpl_propertylist       *paflist;
    cpl_error_code          error;

    cpl_propertylist * pro_tbl = cpl_propertylist_new();

    cpl_propertylist_append_string(pro_tbl,
                                   CPL_DFS_PRO_CATG, procatg_tbl);

    cpl_propertylist_append(pro_tbl, qclist[0]);

    /*******************************/
    /*  Write the FREQ TABLE  */

    /*******************************/

    if(detmon_pernoise_config.mode != 1) {
        /* Set the file name for the table */
        if(!flag_sets) {
            name_o = cpl_sprintf("%s_freq_table.fits", recipe_name);
            assert(name_o != NULL);
        } else {
            name_o =
                cpl_sprintf("%s_freq_table_set%02d.fits", recipe_name,
                            which_set);
            assert(name_o != NULL);
        }

        /* Save the table */
        if(cpl_dfs_save_table(frameset, NULL, parlist, usedframes, NULL, freq_table[0],
                              NULL, recipe_name, pro_tbl, NULL,
                              package, name_o)) {
            cpl_msg_error(cpl_func, "Cannot save the product: %s", name_o);
            cpl_free(name_o);
            cpl_ensure_code(0, CPL_ERROR_FILE_NOT_CREATED);
        }

        if(detmon_pernoise_config.exts < 0) {

            for(i = 1; i < detmon_pernoise_config.nb_extensions; i++) {
                error =
                    cpl_table_save(freq_table[i], NULL, qclist[i], name_o,
                                   CPL_IO_EXTEND);
                cpl_ensure_code(!error, error);
            }
        }

        /* Free */
        cpl_free(name_o);

    } else {
        for (int j = 1; j <= 4; j++) {
            /* Set the file name for the table */
            if(!flag_sets) {
                name_o = cpl_sprintf("%s_freq_table_quad%02d.fits",
                                     recipe_name, j);
                assert(name_o != NULL);
            } else {
                name_o =
                    cpl_sprintf("%s_freq_table_quad%02d_set%02d.fits",
                                recipe_name, j, which_set);
                assert(name_o != NULL);
            }

            /* Save the table */
            if(cpl_dfs_save_table(frameset, NULL, parlist, usedframes, NULL,
                                  freq_table[j - 1],
                                  NULL, recipe_name, pro_tbl, NULL,
                                  package, name_o)) {
                cpl_msg_error(cpl_func, "Cannot save the product: %s", name_o);
                cpl_free(name_o);
                cpl_ensure_code(0, CPL_ERROR_FILE_NOT_CREATED);
            }

            if(detmon_pernoise_config.exts < 0) {
                for(i = 1; i < detmon_pernoise_config.nb_extensions; i++) {
                    error = cpl_table_save(freq_table[(j-1) + 4 * i],
                                           NULL, qclist[i], name_o,
                                           CPL_IO_EXTEND);
                    cpl_ensure_code(!error, error);
                }
            }

            /* Free */
            cpl_free(name_o);
        }

    }
    /*******************************/
    /*  Write the PAF file(s)      */
    /*******************************/

    /* Get FITS header from reference file */
    ref_frame = cpl_frameset_get_position(frameset, 0);
    if((plist = cpl_propertylist_load(cpl_frame_get_filename(ref_frame),
                                      0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_ensure_code(0, cpl_error_get_code());
    }

    /* Get the keywords for the paf file */
    paflist = cpl_propertylist_new();
    cpl_propertylist_copy_property_regexp(paflist, plist,
                                          "^(ARCFILE|MJD-OBS|ESO TPL ID|"
                                          "DATE-OBS|ESO DET DIT|ESO DET NDIT|"
                                          "ESO DET NCORRS|"
                                          "ESO DET MODE NAME)$", 0);

    for(i = 0; i < detmon_pernoise_config.nb_extensions; i++) {
        cpl_propertylist * c_paflist = cpl_propertylist_duplicate(paflist);
        error = cpl_propertylist_append(c_paflist, qclist[i]);
        cpl_ensure_code(!error, error);

        /* Set the file name for the bpm */
        if(detmon_pernoise_config.exts >= 0) {
            if(!flag_sets) {
                name_o = cpl_sprintf("%s.paf", recipe_name);
                assert(name_o != NULL);
            } else {
                name_o = cpl_sprintf("%s_set%02d.paf", recipe_name, which_set);
                assert(name_o != NULL);
            }
        } else {
            if(!flag_sets) {
                name_o = cpl_sprintf("%s_ext%02d.paf", recipe_name, i+1);
                assert(name_o != NULL);
            } else {
                name_o = cpl_sprintf("%s_set%02d_ext%02d.paf", recipe_name, which_set, i+1);
                assert(name_o != NULL);
            }
        }
        /* Save the PAF */
        if(cpl_dfs_save_paf(pipeline_name, recipe_name, c_paflist, name_o)) {
            cpl_msg_error(cpl_func, "Cannot save the product: %s", name_o);
            cpl_free(name_o);
            cpl_propertylist_delete(paflist);
            cpl_propertylist_delete(plist);
            cpl_free(name_o);
            cpl_ensure_code(0, CPL_ERROR_FILE_NOT_CREATED);
        }
        cpl_propertylist_delete(c_paflist);
        cpl_free(name_o);
    }

    cpl_propertylist_delete(plist);
    cpl_propertylist_delete(paflist);
    cpl_propertylist_delete(pro_tbl);
    return cpl_error_get_code();
}
/*---------------------------------------------------------------------------*/
/*
 * @brief  Computes QC for pernoise recipe

 * @param  qclist               computed QC parameters
 * @param  table                input table
 * @param  iquad                input frame quadrant id

 * @return CPL_ERROR_NONE on success.
 */
/*---------------------------------------------------------------------------*/

static cpl_error_code
detmon_pernoise_qc(cpl_propertylist * qclist,
                   cpl_table * table,
                   int iquad)
{
    cpl_error_code error;
    char * propname;

    double freqs[3] = {0, 0, 0};
    double pows[3] = {0, 0, 0};

    /*    error = cpl_propertylist_append_bool(reflist, "POW", TRUE);
          cpl_ensure_code(!error, error);

          error = cpl_table_sort(table, reflist);
          cpl_ensure_code(!error, error);
          */

    int nrows = cpl_table_get_nrow(table);
    int i;

    double * all_freqs = cpl_table_get_data_double(table, "FREQ");
    double * all_pows  = cpl_table_get_data_double(table, "POW");

    for ( i= 1; i< nrows-1; i++){
        if (all_pows[i] > pows[0]) {
            if(all_pows[i-1] < all_pows[i] && all_pows[i] > all_pows[i+1]){
                pows[2]=pows[1];
                pows[1]=pows[0];
                pows[0]=all_pows[i];

                freqs[2]=freqs[1];
                freqs[1]=freqs[0];
                freqs[0]=all_freqs[i];
            }
        } else if (all_pows[i] > pows[1]) {
            if(all_pows[i-1] < all_pows[i] && all_pows[i] > all_pows[i+1]){
                pows[2]=pows[1];
                pows[1]=all_pows[i];

                freqs[2]=freqs[1];
                freqs[1]=all_freqs[i];
            }

        } else if(all_pows[i] > pows[2]) {
            if(all_pows[i-1] < all_pows[i] && all_pows[i] > all_pows[i+1]){
                pows[2]=all_pows[i];

                freqs[2]=all_freqs[i];
            }

        }
    }

    if (detmon_pernoise_config.mode == 1) {
        propname = cpl_sprintf("ESO QC FREQ1 %d", iquad);
        assert(propname != NULL);
    } else {
        propname = cpl_sprintf("ESO QC FREQ1");
    }

    error = cpl_propertylist_append_double(qclist, propname, freqs[0]);
    error = cpl_propertylist_set_comment(qclist,propname,DETMON_QC_FREQ_C);
    cpl_ensure_code(!error, error);

    cpl_free(propname);

    if (detmon_pernoise_config.mode == 1) {
        propname = cpl_sprintf("ESO QC FREQ2 %d", iquad);
        assert(propname != NULL);
    } else {
        propname = cpl_sprintf("ESO QC FREQ2");
    }

    error = cpl_propertylist_append_double(qclist, propname, freqs[1]);
    error = cpl_propertylist_set_comment(qclist,propname,DETMON_QC_FREQ_C);
    cpl_ensure_code(!error, error);

    cpl_free(propname);

    if (detmon_pernoise_config.mode == 1) {
        propname = cpl_sprintf("ESO QC FREQ3 %d", iquad);
        assert(propname != NULL);
    } else {
        propname = cpl_sprintf("ESO QC FREQ3");
    }

    error = cpl_propertylist_append_double(qclist, propname, freqs[2]);
    error = cpl_propertylist_set_comment(qclist,propname,DETMON_QC_FREQ_C);
    cpl_ensure_code(!error, error);

    cpl_free(propname);

    if (detmon_pernoise_config.mode == 1) {
        propname = cpl_sprintf("ESO QC POW1 %d", iquad);
        assert(propname != NULL);
    } else {
        propname = cpl_sprintf("ESO QC POW1");
    }

    error = cpl_propertylist_append_double(qclist, propname, pows[0]);
    error = cpl_propertylist_set_comment(qclist,propname,DETMON_QC_POW_C);
    cpl_ensure_code(!error, error);

    cpl_free(propname);

    if (detmon_pernoise_config.mode == 1) {
        propname = cpl_sprintf("ESO QC POW2 %d", iquad);
        assert(propname != NULL);
    } else {
        propname = cpl_sprintf("ESO QC POW2");
    }

    error = cpl_propertylist_append_double(qclist, propname, pows[1]);
    error = cpl_propertylist_set_comment(qclist,propname,DETMON_QC_POW_C);
    cpl_ensure_code(!error, error);

    cpl_free(propname);

    if (detmon_pernoise_config.mode == 1) {
        propname = cpl_sprintf("ESO QC POW3 %d", iquad);
        assert(propname != NULL);
    } else {
        propname = cpl_sprintf("ESO QC POW3");
    }

    error = cpl_propertylist_append_double(qclist, propname, pows[2]);
    error = cpl_propertylist_set_comment(qclist,propname,DETMON_QC_POW_C);
    cpl_ensure_code(!error, error);


    cpl_free(propname);

    return cpl_error_get_code();
}

/*---------------------------------------------------------------------------*/

/*
 * @brief  Remove background
 * @param  image        Input image
 * @param  nsamples     Input number of (random)samples
 * @param  nffts        input number of FFTs
 * @return CPL_ERROR_NONE on success.
 */

/*---------------------------------------------------------------------------*/
cpl_error_code
detmon_pernoise_rm_bg(cpl_image * image, int nsamples, int nffts)
{
    cpl_vector             *values = cpl_vector_new(nsamples * nffts);

    int                     rejected;
    int i, j;
    cpl_vector *xy_pos = cpl_vector_new(nsamples * nffts * 2);
    cpl_polynomial * poly_2d = 0;
    cpl_image * poly_ima = 0;
    cpl_size degree = 3;
    cpl_error_code error = CPL_ERROR_NONE;
    cpl_matrix * samppos = 0;

    for(i = 1; i <= nffts; i++) {
        for(j = 1; j <= nsamples; j++) {
            cpl_vector_set(xy_pos, (i - 1) * nsamples + (j - 1), j);
            cpl_vector_set(xy_pos, (i - 1) * nsamples + (j - 1) + nsamples * nffts, i);
            cpl_vector_set(values, (i - 1) * nsamples + (j - 1),
                           cpl_image_get(image, j, i, &rejected));
            error = cpl_error_get_code();
            if (error != CPL_ERROR_NONE)
            {
                break;
            }
        }
        if (error != CPL_ERROR_NONE)
        {
            break;
        }
    }
    if (error != CPL_ERROR_NONE)
    {
        goto cleanup;
    }

    poly_2d = cpl_polynomial_new(2);
    samppos =
        cpl_matrix_wrap(2, nsamples * nffts, cpl_vector_get_data(xy_pos));

    cpl_polynomial_fit(poly_2d, samppos, NULL, values, NULL,
                       CPL_FALSE, NULL, &degree);

    cpl_matrix_unwrap(samppos);

    poly_ima = cpl_image_new(nsamples, nffts, CPL_TYPE_FLOAT);

    cpl_image_fill_polynomial(poly_ima, poly_2d, 1, 1, 1, 1);

    cpl_image_subtract(image, poly_ima);

cleanup:
    cpl_polynomial_delete(poly_2d);
    cpl_image_delete(poly_ima);
    cpl_vector_delete(xy_pos);
    cpl_vector_delete(values);

    return cpl_error_get_code();
}

/**@}*/
